module m_analisis_dinamico

! ::: MÓDULO: m_analisis_dinamico ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!      Este módulo recoge las funciones para el análisis numérico de las ecuaciones diferenciales del modelo de Hodgkin-Huxley reducido. Se
!   divide en análisis de puntos de equilibrio, análisis de bifurcación de Hopf y análisis de bifurcación SN.
!
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

use m_parametros
use m_modelo_HH_2D
use m_subrutinas

public  :: pto_equilibrio, info_pto_equilibrio, &
           bifurcacion_Hopf, traza_J, &
           I_Vb, dv_I_Vb, ddv_I_Vb, db_I_Vb, b_V, d_b_V, bifurcacion_SN
private :: calculos_aux, f_xx, f_yy, f_xy, f_xxx, f_xyy, g_xx, g_yy, g_xy, g_xxy, g_yyy

contains

! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: PUNTOS DE EQUILIBRIO

subroutine pto_equilibrio(punto_eq,tr_J,det_J,vp_r,vp_i,precision)

! ::    Subrutina de cálculo de puntos de equilibrio basada en los métodos de Newton-Raphson y Gauss-Jordan.
! :: 
! :: Argumentos:
! ::    - punto_eq (Vector de componentes reales): Recoge la primera aproximación al punto de equilibrio. Los valores proporcionados por
! ::                                               el usuario son sustituidos con el resultado exacto.
! ::    - tr_J, det_J (Variables reales): La subrutina asigna a estas variables los valores de la traza y el determinante del jacobiano,
! ::                                      respectivamente, evaluado en el punto de equilibrio.
! ::    - vp_r, vp_i (Vectores de componentes reales): La subrutina asigna a estas variables la parte real e imaginaria de los valores 
! ::                                                   propios, respectivamente.
! ::    - precision (Opcional, variable real): Especifica la precisión a usar (valor predeterminado: PRECISION_DP).

     real(kind=dp), dimension(:), intent(inout)            :: punto_eq
     real(kind=dp), intent(out)                            :: tr_J, det_J
     real(kind=dp), dimension(size(punto_eq)), intent(out) :: vp_r, vp_i
     real(kind=dp), intent(in), optional                   :: precision

     real(kind=dp), dimension(2)                           :: y_temp
     real(kind=dp), dimension(2,2)                         :: matriz_A
     real(kind=dp)                                         :: t_aux, PRECISION_BUSQUEDA_PTO

     real(kind=dp), dimension(2)                           :: matriz_C

     if (present(precision)) then
          PRECISION_BUSQUEDA_PTO = precision
     else
          PRECISION_BUSQUEDA_PTO = PRECISION_DP
     end if

     do ! Búsqueda del punto de equilibrio (Newton-Raphson + Gauss-Jordan)
          matriz_A = HH_jacobiano_2D(punto_eq)      ! Matriz de coeficientes
          matriz_C = - HH_modelo_2D(t_aux,punto_eq) ! Términos independientes

          call gauss_jordan(matriz_A,matriz_C)
          if ( abs(matriz_C(1)) < PRECISION_BUSQUEDA_PTO .or. abs(matriz_C(2)) < PRECISION_BUSQUEDA_PTO) then
               exit
          end if

          punto_eq(1) = punto_eq(1) + matriz_C(1)
          punto_eq(2) = punto_eq(2) + matriz_C(2)
     end do

     ! ----- ESTABILIDAD: -----
     matriz_A = HH_jacobiano_2D(punto_eq)

     tr_J = matriz_A(1,1) + matriz_A(2,2)
     det_J = matriz_A(1,1) * matriz_A(2,2) - matriz_A(1,2) * matriz_A(2,1)

     call eival(matriz_A,vp_r,vp_i)

end subroutine pto_equilibrio

subroutine info_pto_equilibrio(tr_J,det_J,tipo_pto_eq)

! ::    Subrutina de análisis del punto de equilibrio. A través de los valores proporcionados de la traza y determinante del jacobiano 
! :: (evaluados en el punto de equilibrio) se determina el tipo de punto, siguiendo la clasificación del diagrama de Poincaré (fig. 3.1).
! :: 
! :: Argumentos:
! ::    - tr_J, det_J (Variables reales): Valores de la traza y determinante del jacobiano proporcionados por el usuario.
! ::    - tipo_pto_eq (Variable entera): La subrutina asigna a esta variable un número entero según el tipo de punto:
! ::          1: Nodo estable
! ::          2: Nodo inestable
! ::          3: Foco estable
! ::          4: Foco intestable
! ::          5: Punto de silla
! ::          6: Punto en bifurcación
! ::         -1: Caso degenerado (El punto se encuentra sobre la curva τ^2-4∆=0)

     real(kind=dp), intent(in) :: tr_J, det_J
     integer, intent(out)      :: tipo_pto_eq

     real(kind=dp)             :: zero_sqrt

     zero_sqrt = (tr_J)**2 - 4.0_dp * det_J

     if (abs(det_J) < 0.00001_dp) then
          tipo_pto_eq = 6
     else if (det_J < 0.0_dp) then
          tipo_pto_eq = 5
     else
          if (abs(tr_J) < 0.00001_dp) then
               tipo_pto_eq = 6
          else if (tr_J < 0.0_dp) then
               if (abs(zero_sqrt) < 0.00001_dp) then
                    tipo_pto_eq = -1
               else if (zero_sqrt < 0.0_dp) then
                    tipo_pto_eq = 3
               else
                    tipo_pto_eq = 1
               end if
          else
               if (abs(zero_sqrt) < 0.00001_dp) then
                    tipo_pto_eq = -1
               else if (zero_sqrt < 0.0_dp) then
                    tipo_pto_eq = 4
               else
                    tipo_pto_eq = 2
               end if 
          end if
     end if

end subroutine info_pto_equilibrio

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::



! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: ANÁLISIS BIFURCACIÓN DE HOPF

subroutine calculos_aux(punto_eq,F_v,F_n,w,x,y)

     real(kind=dp), dimension(:), intent(in)                 :: punto_eq
     real(kind=dp), intent(out)                              :: F_v, F_n, w, x, y

     real(kind=dp)                                           :: tr_J, det_J
     real(kind=dp), dimension(size(punto_eq))                :: temp, vp_r, vp_i
     real(kind=dp), dimension(size(punto_eq),size(punto_eq)) :: matriz_J

     temp = punto_eq
     call pto_equilibrio(temp,tr_J,det_J,vp_r,vp_i)

     matriz_J = HH_jacobiano_2D(temp)
     F_v = matriz_J(1,1)
     F_n = matriz_J(1,2)
     w = sqrt(det_J)
     x = temp(1)
     y = - (F_n * punto_eq(2) + F_v * x) / w

end subroutine calculos_aux

function f_xx(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxx, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxx = (1.0_dp / C_mem) * ( - g_Na * (  3.0_dp * m_inf_2D(punto_eq) * (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) &
                                                 * (x - E_Na) * (2.0_dp * (d_m_inf_2D(punto_eq))**2 &
                                                 + m_inf_2D(punto_eq) * d2_m_inf_2D(punto_eq)) &
                                            + 6.0_dp * (m_inf_2D(punto_eq))**2 * d_m_inf_2D(punto_eq) * ( 1.0096_dp * (F_v / F_n) &
                                                 * (x - E_Na) + (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) ) &
                                            + 2.0192_dp * (F_v / F_n) * (m_inf_2D(punto_eq))**3 &
                                         ) &
                                - 4.0_dp * g_K * (F_v / F_n) * ((F_v * x + w * y) / F_n)**2 * ( 3.0_dp * (F_v / F_n) * (x - E_K) &
                                                                                            + 2.0_dp * ((F_v * x + w * y) / F_n) ) &
                              )

     d = Fxx

end function f_xx

function f_yy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fyy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fyy = - (1.0_dp / C_mem) * 12.0_dp * g_K * (x - E_K) * ((F_v * x + w * y) / F_n)**2 * (w / F_n)**2

     d = Fyy

end function f_yy

function f_xy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxy = (1.0_dp / C_mem) * (  - g_Na * 1.0096_dp * (m_inf_2D(punto_eq))**2 * (w / F_n)**2 &
                                      * (  3.0_dp * d_m_inf_2D(punto_eq) * (x - E_Na) + m_inf_2D(punto_eq)  ) &
                                 - g_K * 4.0_dp * ((F_v * x + w * y) / F_n)**2 * (w / F_n) &
                                      * (  3.0_dp * (F_v / F_n) * (x - E_K) + ((F_v * x + w * y) / F_n)  ) &
                              )

     d = Fxy

end function f_xy

function f_xxx(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxxx, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxxx = (1.0_dp / C_mem) * (   - g_Na * (  6.0_dp * (d_m_inf_2D(punto_eq))**3 &
                                                  * (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) * (x - E_Na) &
                                             + 18.0_dp * m_inf_2D(punto_eq) * (d_m_inf_2D(punto_eq))**2 &
                                                  * ( 1.0096_dp * (F_v / F_n) * (x - E_Na) &
                                                  + (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) ) &
                                             + 18.0_dp * m_inf_2D(punto_eq) * d_m_inf_2D(punto_eq) * d2_m_inf_2D(punto_eq) & 
                                                  * (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) * (x - E_Na) &
                                             + 3.0_dp * (m_inf_2D(punto_eq))**2 * d3_m_inf_2D(punto_eq) &
                                                  * (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) * (x - E_Na) &
                                             + 9.0_dp * (m_inf_2D(punto_eq))**2 * d2_m_inf_2D(punto_eq) &
                                                  * ( 1.0096_dp * (F_v / F_n) * (x - E_Na) &
                                                  + (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) ) &
                                             + 6.0_dp * (m_inf_2D(punto_eq))**2 * d_m_inf_2D(punto_eq) &
                                                  * (1.0_dp + 2.0192_dp * (F_v / F_n)) &
                                            ) &
                                   - g_K * 12.0_dp * (F_v / F_n) * ((F_v * x + w * y) / F_n) &
                                        * ( 2.0_dp * (F_v / F_n)**2 * (x - E_K) &
                                        + ((F_v * x + w * y) / F_n) * (2.0_dp + (F_v / F_n)) ) &
                               )

     d = Fxxx

end function f_xxx

function f_xyy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxyy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxyy = - (1.0_dp / C_mem) * 12.0_dp * g_K * (w / F_n)**2 * ( (F_v * x + w * y) / F_n ) &
            * ( 2.0_dp * (F_v / F_n) * (x - E_Na) + ((F_v * x + w * y) / F_n) )

     d = Fxyy

end function f_xyy

function g_xx(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxx, Gxx, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxx = (1.0_dp / C_mem) * ( - g_Na * (  3.0_dp * m_inf_2D(punto_eq) * (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) &
                                               * (x - E_Na) * (2.0_dp * (d_m_inf_2D(punto_eq))**2 + m_inf_2D(punto_eq) &
                                               * d2_m_inf_2D(punto_eq)) &
                                          + 6.0_dp * (m_inf_2D(punto_eq))**2 * d_m_inf_2D(punto_eq) * ( 1.0096_dp * (F_v / F_n) &
                                               * (x - E_Na) + (0.9088_dp + 1.0096_dp * ((F_v * x + w * y) / F_n)) ) &
                                          + 2.0192_dp * (F_v / F_n) * (m_inf_2D(punto_eq))**3 &
                                         ) &
                                - 4.0_dp * g_K * (F_v / F_n) * ((F_v * x + w * y) / F_n)**2 * ( 3.0_dp * (F_v / F_n) * (x - E_K) &
                                                                                            + 2.0_dp * ((F_v * x + w * y) / F_n) ) &
                              )

     Gxx = (  F_n * tau_n_2D(punto_eq) * ( (F_n)**2 * d2_n_inf_2D(punto_eq) * tau_n_2D(punto_eq) &
                 - (F_n * n_inf_2D(punto_eq) + (F_v * x + w * y)) * F_n * d2_tau_n_2D(punto_eq) ) &
            - 2.0_dp * d_tau_n_2D(punto_eq) * ( (F_n * d_n_inf_2D(punto_eq) + F_v) * F_n * tau_n_2D(punto_eq) &
                 - (F_n * n_inf_2D(punto_eq) + (F_v * x + w * y)) * F_n * d_tau_n_2D(punto_eq) )  ) &
         / ( F_n * tau_n_2D(punto_eq) )**3

     d = - (1.0_dp / w) * ( F_v * Fxx + F_n * Gxx )

end function g_xx

function g_yy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fyy, Gyy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fyy = - (1.0_dp / C_mem) * 12.0_dp * g_K * (x - E_K) * ((F_v * x + w * y) / F_n)**2 * (w / F_n)**2

     Gyy = 0.0_dp

     d = - (1.0_dp / w) * ( F_v * Fyy + F_n * Gyy )

end function g_yy

function g_xy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxy, Gxy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxy = (1.0_dp / C_mem) * (  - g_Na * 1.0096_dp * (m_inf_2D(punto_eq))**2 * (w / F_n)**2 &
                                      * (  3.0_dp * d_m_inf_2D(punto_eq) * (x - E_Na) + m_inf_2D(punto_eq)  ) &
                                 - g_K * 4.0_dp * ((F_v * x + w * y) / F_n)**2 * (w / F_n) &
                                      * (  3.0_dp * (F_v / F_n) * (x - E_K) + ((F_v * x + w * y) / F_n)  ) &
                              )

     Gxy = - ( w * d_tau_n_2D(punto_eq) ) / ( F_n * (tau_n_2D(punto_eq))**2 )

     d = - (1.0_dp / w) * ( F_v * Fxy + F_n * Gxy )

end function g_xy

function g_xxy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fxxy, Gxxy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fxxy = (1.0_dp / C_mem) * ( - g_Na * 3.0288_dp * (w / F_n) * (  2.0_dp * m_inf_2D(punto_eq) * (d_m_inf_2D(punto_eq))**2 &
                                                                        * (x - E_Na) &
                                                                   + (m_inf_2D(punto_eq))**2 * d2_m_inf_2D(punto_eq) * (x - E_Na) &
                                                                   + 2.0_dp * (m_inf_2D(punto_eq))**2 * d_m_inf_2D(punto_eq) &
                                                                  ) &
                                 - g_K * 24.0_dp * ((w * F_v) / (F_n)**4) * (F_v * x + w * y) &
                                      * ( F_v * (x - E_K) + (F_v * x + w * y) ) &
                               )

     Gxxy = ( w / (F_n * (tau_n_2D(punto_eq))**2) ) &
            * ( ( (2.0_dp * (d_tau_n_2D(punto_eq))**2) / (F_n * tau_n_2D(punto_eq)) ) - d2_tau_n_2D(punto_eq) )

     d = - (1.0_dp / w) * ( F_v * Fxxy + F_n * Gxxy )

end function g_xxy

function g_yyy(punto_eq) result(d)

     real(kind=dp), dimension(:), intent(in) :: punto_eq
     real(kind=dp)                           :: d

     real(kind=dp)                           :: Fyyy, Gyyy, F_v, F_n, w, x, y

     call calculos_aux(punto_eq,F_v,F_n,w,x,y)

     Fyyy = - ( 24.0_dp * g_K * (x - E_K) * (F_v * x + w * y) ) / ( C_mem * (F_n)**4 )

     Gyyy = 0.0_dp

     d = - (1.0_dp / w) * ( F_v * Fyyy + F_n * Gyyy )

end function g_yyy





subroutine bifurcacion_Hopf(punto_eq,b,tr_J,det_J,a)

     real(kind=dp), dimension(:), intent(inout) :: punto_eq
     real(kind=dp), intent(in)                  :: b
     real(kind=dp), intent(out)                 :: tr_J, det_J, a

     real(kind=dp), dimension(size(punto_eq))   :: temp1, temp2

     E_K = b

     ! No hiperbolicidad:
     call pto_equilibrio(punto_eq,tr_J,det_J,temp1,temp2)

     ! No degeneración:
     a =   ( 1.0_dp / 16.0_dp ) * ( f_xxx(punto_eq) + f_xyy(punto_eq) + g_xxy(punto_eq) + g_yyy(punto_eq) ) &
         + (  1.0_dp / ( 16.0_dp * sqrt(det_J) )  ) * (  f_xy(punto_eq) * ( f_xx(punto_eq) + f_yy(punto_eq) ) &
                                                       - g_xy(punto_eq) * ( g_xx(punto_eq) + g_yy(punto_eq) ) &
                                                       - f_xx(punto_eq) * g_xx(punto_eq) + f_yy(punto_eq) * g_yy(punto_eq)  )

end subroutine bifurcacion_Hopf

function traza_J(b) result(tau)

     real(kind=dp), intent(in)     :: b
     real(kind=dp)                 :: tau

     E_K = b

     call pto_equilibrio(pto_eq,tau,null_float,null_farray2,null_farray2)

end function traza_J

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::



! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: ANÁLISIS BIFURCACIÓN SN

function I_Vb(V) result(x)

     real(kind=dp), intent(in)   :: V
     real(kind=dp)               :: x

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)

     x = (1.0_dp / C_mem) * ( I_ext - g_Na * ( m_inf_2D(y) )**3 * ( 0.9088_dp - 1.0096_dp * n_inf_2D(y) ) * (V - E_Na)  &
                                      - g_K * ( n_inf_2D(y) )**4 * (V - E_K) - g_L * (V - E_L) )

end function I_Vb

function dv_I_Vb(V) result(x)

     real(kind=dp), intent(in)   :: V
     real(kind=dp)               :: x

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)

     x = (1.0_dp / C_mem) * (  - g_Na * ( 3.0_dp * (m_inf_2D(y))**2 * d_m_inf_2D(y) * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                             * (V - E_Na) &
                                        - 1.0096_dp * (m_inf_2D(y))**3 * d_n_inf_2D(y) * (V - E_Na) &
                                        + (m_inf_2D(y))**3 * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                        ) &
                               - g_K * ( 4.0_dp * (n_inf_2D(y))**3 * d_n_inf_2D(y) * (V - E_K) + (n_inf_2D(y))**4 ) &
                               - g_L &
                            )

end function dv_I_Vb

function ddv_I_Vb(V) result(x)

     real(kind=dp), intent(in)   :: V
     real(kind=dp)               :: x

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)

     x = (1.0_dp / C_mem) * (  - g_Na * ( 6.0_dp * m_inf_2D(y) * (d_m_inf_2D(y))**2 * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                             * (V - E_Na) &
                                        + 3.0_dp * (n_inf_2D(y))**2 * d2_m_inf_2D(y) * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                             * (V - E_Na) &
                                        + 6.0_dp * (n_inf_2D(y))**2 * d_m_inf_2D(y) * ( (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                             - 1.0096_dp * d_n_inf_2D(y) * (V - E_Na) ) &
                                        - 1.0096_dp * (n_inf_2D(y))**3 * ( d2_n_inf_2D(y) * (V - E_Na) + 2.0_dp * d_n_inf_2D(y) ) &
                                        ) &
                               - g_K * ( 12.0_dp * (n_inf_2D(y))**2 * (d_n_inf_2D(y))**2 * (V - E_K) &
                                       + 4.0_dp * (n_inf_2D(y))**3 * d2_n_inf_2D(y) * (V - E_K) &
                                       + 8.0_dp * (n_inf_2D(y))**3 * d_n_inf_2D(y) &
                                       ) &
                            )

end function ddv_I_Vb

function db_I_Vb(V) result(x)

     real(kind=dp), intent(in)   :: V
     real(kind=dp)               :: x

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)

     x = (1.0_dp / C_mem) * g_K * (n_inf_2D(y))**4

end function db_I_Vb

function b_V(V) result(x)

     real(kind=dp), intent(in)   :: V
     real(kind=dp)               :: x

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)

     x = V - (  ( I_ext - g_Na * (m_inf_2D(y))**3 * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) * (V - E_Na) - g_L * (V - E_L) ) &
         / ( g_K * (n_inf_2D(y))**4 )  )

end function b_V

function d_b_V(V) result(x)

     real(kind=dp), intent(in)   :: V
     real(kind=dp)               :: x

     real(kind=dp), dimension(2) :: y
     real(kind=dp)               :: num1, num2, den

     y = (/V, null_float/)

     num1 = g_K * (n_inf_2D(y))**4 * (  - g_Na * ( 3.0_dp * (m_inf_2D(y))**2 * d_m_inf_2D(y) &
                                                   * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) * (V - E_Na) &
                                              - 1.0096_dp * (m_inf_2D(y))**3 * d_n_inf_2D(y) * (V - E_Na) &
                                              + (m_inf_2D(y))**3 * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                              ) &
                                     - g_L &
                                  )
     num2 = 4.0_dp * g_K * (n_inf_2D(y))**3 * d_n_inf_2D(y) * ( I_ext &
                                                        - g_Na * (m_inf_2D(y))**3 * (0.9088_dp - 1.0096_dp * n_inf_2D(y)) &
                                                             * (V - E_Na) &
                                                        - g_L * (V - E_L) &
                                                        )
     den = (  g_K * (n_inf_2D(y))**4  )**2

     x = 1.0_dp - (  ( num1 - num2 ) / den  )

end function d_b_V





subroutine bifurcacion_SN(punto_eq,b,det_J,dv_I,a,c)

     real(kind=dp), dimension(:), intent(inout) :: punto_eq
     real(kind=dp), intent(in)                  :: b
     real(kind=dp), intent(out)                 :: det_J, dv_I, a, c

     real(kind=dp)                              :: tr_J
     real(kind=dp), dimension(size(punto_eq))   :: temp1, temp2

     E_K = b

     ! No hiperbolicidad:
     call pto_equilibrio(punto_eq,tr_J,det_J,temp1,temp2)
     dv_I = dv_I_Vb(punto_eq(1))

     ! No degeneración:
     a = 0.5_dp * ddv_I_Vb(punto_eq(1))

     ! Transversalidad:
     c = db_I_Vb(punto_eq(1))

end subroutine bifurcacion_SN

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

end module m_analisis_dinamico
