module m_modelo_HH_4D

! ::: MÓDULO: m_modelo_HH_4D :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!      Este módulo recoge las funciones necesarias para definir las ecuaciones diferenciales que forman el modelo de Hodgkin-Huxley 4D, así como
!   las funciones de activación/inactivación, las constantes de tiempo y la función que calcula las condiciones iniciales.
!
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

use m_parametros

public  :: HH_funciones_act_inact_4D, HH_ctes_tiempo_4D, &
           HH_modelo_4D, HH_cond_iniciales_4D
private :: m_inf_4D, tau_m_4D, h_inf_4D, tau_h_4D, n_inf_4D, tau_n_4D

contains

! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: FUNCIONES SECUNDARIAS

function m_inf_4D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = 1.0_dp / ( 1.0_dp + exp((V_half(1) - y(1)) / k_slope(1)) )

end function m_inf_4D

function tau_m_4D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = C_base(1) + C_amp(1) * exp(  ( - (V_max(1) - y(1))**2 ) / ( sigma(1) * sigma(1) )  )

end function tau_m_4D

function h_inf_4D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = 1.0_dp / ( 1.0_dp + exp((V_half(2) - y(1)) / k_slope(2)) )

end function h_inf_4D

function tau_h_4D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = C_base(2) + C_amp(2) * exp(  ( - (V_max(2) - y(1))**2 ) / ( sigma(2) * sigma(2) )  )

end function tau_h_4D

function n_inf_4D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = 1.0_dp / ( 1.0_dp + exp((V_half(3) - y(1)) / k_slope(3)) )

end function n_inf_4D

function tau_n_4D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = C_base(3) + C_amp(3) * exp(  ( - (V_max(3) - y(1))**2 ) / ( sigma(3) * sigma(3) )  )

end function tau_n_4D

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: FUNCIONES DE ACTIVACIÓN/INACTIVACIÓN Y CONSTANTES DE TIEMPO

subroutine HH_funciones_act_inact_4D(V_mem,y)

     real(kind=dp), intent(in)                  :: V_mem
     real(kind=dp), dimension(:), intent(inout) :: y

     y(1) = V_mem
     y(2) = m_inf_4D(y)
     y(3) = h_inf_4D(y)
     y(4) = n_inf_4D(y)

end subroutine HH_funciones_act_inact_4D

subroutine HH_ctes_tiempo_4D(V_mem,y)

     real(kind=dp), intent(in)                  :: V_mem
     real(kind=dp), dimension(:), intent(inout) :: y

     y(1) = V_mem
     y(2) = tau_m_4D(y)
     y(3) = tau_h_4D(y)
     y(4) = tau_n_4D(y)

end subroutine HH_ctes_tiempo_4D

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: MODELO DE HODGKIN-HUXLEY (4D)

function HH_modelo_4D(t,y) result(yprima)

     real(kind=dp), intent(in)               :: t
     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp), dimension(size(y))       :: yprima

     yprima(1) = (1.0_dp / C_mem) * ( I_ext - g_Na * (y(2))**3 * y(3) * (y(1) - E_Na)  - g_K * (y(4))**4 * (y(1) - E_K) &
                                    - g_L * (y(1) - E_L) )
     yprima(2) = (m_inf_4D(y) - y(2)) / tau_m_4D(y)
     yprima(3) = (h_inf_4D(y) - y(3)) / tau_h_4D(y)
     yprima(4) = (n_inf_4D(y) - y(4)) / tau_n_4D(y)

end function HH_modelo_4D

subroutine HH_cond_iniciales_4D(E_K_find,y)

     real(kind=dp), intent(in)                :: E_K_find
     real(kind=dp), dimension(:), intent(out) :: y

     real(kind=dp)               :: E_K_file
     real(kind=dp), dimension(4) :: y_old
     real(kind=dp), dimension(2) :: E_K_temp
     integer                     :: iostat_read

     E_K_temp(1) = E_K_find * 1000.0_dp
     E_K_temp(1) = anint(E_K_temp(1))

     open(unit=1,file=".COND_INICIALES_4D.txt",status="old",action="read",position="rewind")
     y = 0.0_dp
     do
          y_old = y
          read(unit=1,fmt=*,iostat=iostat_read) E_K_file, y(1), y(2), y(3), y(4)
          if (iostat_read < 0) then
               exit
          end if
          E_K_temp(2) = E_K_file * 1000.0_dp
          E_K_temp(2) = anint(E_K_temp(2))
          if (E_K_temp(1) == E_K_temp(2)) then
               return
          end if
     end do
     print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
     print *, "*  ERROR HH_COND_INICIALES_4D: No se han encontrado las condiciones iniciales     *"
     print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
     close(unit=1)

end subroutine HH_cond_iniciales_4D

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

end module m_modelo_HH_4D





module m_modelo_HH_2D

! ::: MÓDULO: m_modelo_HH_2D :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!      Este módulo recoge las funciones necesarias para definir las ecuaciones diferenciales que forman el modelo de Hodgkin-Huxley reducido,
!   las funciones de activación/inactivación y la constante de tiempo, la función estímulo y la función que calcula las condiciones iniciales. 
!
!      El análisis dinámico se realiza sobre el modelo reducido. Por tanto, también se incluye la función jacobiano y las funciones
!   necesarias para el cálculo de las isoclinas.
!
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

use m_parametros 
use m_subrutinas

public  :: m_inf_2D, n_inf_2D, tau_n_2D, d_m_inf_2D, d2_m_inf_2D, d3_m_inf_2D, d_n_inf_2D, d2_n_inf_2D, d_tau_n_2D, d2_tau_n_2D, &
           HH_funciones_act_inact_2D, HH_ctes_tiempo_2D, &
           HH_modelo_2D, HH_estimulo, HH_cond_iniciales_2D, HH_jacobiano_2D, &
           HH_isoclina_V_2D, HH_isoclina_V_2D_prima, &
           HH_isoclina_n_2D, HH_isoclina_n_2D_prima
private :: func_aux_2D, func_aux_2D_prima

real(kind=dp), private :: V_aux

contains

! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: FUNCIONES SECUNDARIAS

function m_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = 1.0_dp / ( 1.0_dp + exp((V_half(1) - y(1)) / k_slope(1)) )

end function m_inf_2D

function n_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = 1.0_dp / ( 1.0_dp + exp((V_half(3) - y(1)) / k_slope(3)) )

end function n_inf_2D

function tau_n_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = C_base(3) + C_amp(3) * exp(  ( - (V_max(3) - y(1))**2 ) / ( sigma(3) * sigma(3) )  )

end function tau_n_2D

function d_m_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = exp((V_half(1) - y(1)) / k_slope(1)) / (  k_slope(1) * ( exp((V_half(1) - y(1)) / k_slope(1)) + 1.0_dp )**2  )

end function d_m_inf_2D

function d2_m_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = ( (exp((V_half(1) - y(1)) / k_slope(1)) - 1.0_dp) * exp((V_half(1) - y(1)) / k_slope(1)) ) &
       / ( k_slope(1) * k_slope(1) * (exp((V_half(1) - y(1)) / k_slope(1)) + 1.0_dp)**3 )

end function d2_m_inf_2D

function d3_m_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = (  exp((V_half(1) - y(1)) / k_slope(1)) * ( exp(2.0_dp * (V_half(1) - y(1)) / k_slope(1)) &
                                                    - 4.0_dp * exp((V_half(1) - y(1)) / k_slope(1)) + 1.0_dp )  ) &
       / (  (k_slope(1))**3 * (exp((V_half(1) - y(1)) / k_slope(1)) + 1.0_dp)**4  )

end function d3_m_inf_2D

function d_n_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = exp((V_half(3) - y(1)) / k_slope(3)) / (  k_slope(3) * ( exp((V_half(3) - y(1)) / k_slope(3)) + 1.0_dp )**2  )

end function d_n_inf_2D

function d2_n_inf_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = ( (exp((V_half(3) - y(3)) / k_slope(3)) - 1.0_dp) * exp((V_half(3) - y(3)) / k_slope(3)) ) &
       / ( k_slope(3) * k_slope(3) * (exp((V_half(3) - y(3)) / k_slope(3)) + 1.0_dp)**3 )

end function d2_n_inf_2D

function d_tau_n_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = (  2.0_dp * C_amp(3) * (V_max(3) - y(1)) * exp( - (V_max(3) - y(1))**2 / (sigma(3) * sigma(3)) )  ) &
       / ( sigma(3) * sigma(3) )

end function d_tau_n_2D

function d2_tau_n_2D(y) result(x)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp)                           :: x

     x = (  2.0_dp * C_amp(3) * (2.0_dp * (y(1))**2 - 4.0_dp * V_max(3) * y(1) - (sigma(3))**2 + 2.0_dp * (V_max(3))**2) &
          * exp( - (V_max(3) - y(1))**2 / (sigma(3) * sigma(3)) )  ) &
       / (  sigma(3)  )**4

end function d2_tau_n_2D

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: FUNCIONES DE ACTIVACIÓN/INACTIVACIÓN Y CONSTANTES DE TIEMPO

subroutine HH_funciones_act_inact_2D(V_mem,y)

     real(kind=dp), intent(in)                  :: V_mem
     real(kind=dp), dimension(:), intent(inout) :: y

     y(1) = V_mem
     y(2) = n_inf_2D(y)

end subroutine HH_funciones_act_inact_2D

subroutine HH_ctes_tiempo_2D(V_mem,y)

     real(kind=dp), intent(in)                  :: V_mem
     real(kind=dp), dimension(:), intent(inout) :: y

     y(1) = V_mem
     y(2) = tau_n_2D(y)

end subroutine HH_ctes_tiempo_2D

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: MODELO DE HODGKIN-HUXLEY (2D)

function HH_modelo_2D(t,y) result(yprima)

     real(kind=dp), intent(in)               :: t
     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp), dimension(size(y))       :: yprima

     yprima(1) = (1.0_dp / C_mem) * ( I_ext - g_Na * ( m_inf_2D(y) )**3 * ( 0.9088_dp - 1.0096_dp * y(2) ) * (y(1) - E_Na)  &
                                      - g_K * (y(2))**4 * (y(1) - E_K) - g_L * (y(1) - E_L) )
     yprima(2) = (n_inf_2D(y) - y(2)) / tau_n_2D(y)

end function HH_modelo_2D

subroutine HH_estimulo(t, I_externa, t_inicial, t_final)

     real(kind=dp), intent(in)           :: t, I_externa
     real(kind=dp), intent(in), optional :: t_inicial, t_final

     if (present(t_inicial) .and. present(t_final)) then ! El estímulo se inicia en t_inicial y acaba en t_final
          if (t > t_inicial .and. t < t_final) then
               I_ext = I_externa
          else
               I_ext = 0.0_dp
          end if
     else if (present(t_inicial)) then                   ! El estímulo se inicia en t_inicial y se mantiene hasta el final
          if (t > t_inicial) then
               I_ext = I_externa
          else
               I_ext = 0.0_dp
          end if
     else if (present(t_final)) then                     ! El estímulo está activado desde el principio hasta t_final
          if (t > t_final) then
               I_ext = 0.0_dp
          else
               I_ext = I_externa
          end if
     else                                                ! Estímulo siempre activado
          I_ext = I_externa
     end if

end subroutine HH_estimulo

subroutine HH_cond_iniciales_2D(E_K_find,y)

     real(kind=dp), intent(in)                :: E_K_find
     real(kind=dp), dimension(:), intent(out) :: y

     real(kind=dp)               :: E_K_file
     real(kind=dp), dimension(2) :: y_old, E_K_temp
     integer                     :: iostat_read

     E_K_temp(1) = E_K_find * 1000.0_dp
     E_K_temp(1) = anint(E_K_temp(1))

     open(unit=1,file=".COND_INICIALES_2D.txt",status="old",action="read",position="rewind")
     y = 0.0_dp
     do
          y_old = y
          read(unit=1,fmt=*,iostat=iostat_read) E_K_file, y(1), y(2)
          if (iostat_read < 0) then
               exit
          end if
          E_K_temp(2) = E_K_file * 1000.0_dp
          E_K_temp(2) = anint(E_K_temp(2))
          if (E_K_temp(1) == E_K_temp(2)) then
               return
          end if
     end do      
     print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
     print *, "*  ERROR HH_COND_INICIALES_2D: No se han encontrado las condiciones iniciales     *"
     print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
     close(unit=1)

end subroutine HH_cond_iniciales_2D

function HH_jacobiano_2D(y) result(J)

     real(kind=dp), dimension(:), intent(in) :: y
     real(kind=dp), dimension(2,2)           :: J

     J(1,1) = (1.0_dp / C_mem) * (  - g_Na * 3.0_dp * ( m_inf_2D(y) )**2 * d_m_inf_2D(y) &
                                    * ( 0.9088_dp - 1.0096_dp * y(2) ) * (y(1) - E_Na) &
                                    - g_Na * ( m_inf_2D(y) )**3 * ( 0.9088_dp - 1.0096_dp * y(2) ) &
                                    - g_K * (y(2))**4 - g_L  )
     J(1,2) = (1.0_dp / C_mem) * (  g_Na * 1.0096_dp * ( m_inf_2D(y) )**3 * (y(1) - E_Na) &
                                    - 4.0_dp * g_K * (y(2))**3 * (y(1) - E_K)  )
     J(2,1) = ( d_n_inf_2D(y) * tau_n_2D(y) - (n_inf_2D(y) - y(2)) * d_tau_n_2D(y) ) / ( tau_n_2D(y) * tau_n_2D(y) )
     J(2,2) = - 1.0_dp / tau_n_2D(y)

end function HH_jacobiano_2D



! --- Isoclina cero V: ---

function func_aux_2D(n) result(f_V)

     real(kind=dp), intent(in) :: n
     real(kind=dp)             :: f_V

     real(kind=dp), dimension(2) :: y

     y = (/V_aux, null_float/)
     f_V = I_ext - g_Na * ( m_inf_2D(y) )**3 * ( 0.9088_dp - 1.0096_dp * n ) * (V_aux - E_Na)  &
         - g_K * n**4 * (V_aux - E_K) - g_L * (V_aux - E_L)

end function func_aux_2D

function HH_isoclina_V_2D(V) result(n_func)

     real(kind=dp), intent(in) :: V
     real(kind=dp)             :: n_func

     integer :: iostat_raiz

     V_aux = V

     call biseccion(func_aux_2D,0.0_dp,1.0_dp,n_func,iostat_raiz)

end function HH_isoclina_V_2D

function func_aux_2D_prima(n) result(f_V)

     real(kind=dp), intent(in) :: n
     real(kind=dp)             :: f_V

     real(kind=dp), dimension(2) :: y

     y = (/V_aux, null_float/)
     f_V = - 3.0_dp * g_Na * ( m_inf_2D(y) )**2 * d_m_inf_2D(y) * ( 0.9088_dp - 1.0096_dp * n ) * (V_aux - E_Na) &
           - g_Na * ( m_inf_2D(y) )**3 * ( 0.9088_dp - 1.0096_dp * n ) &
           - g_K * n**4 - g_L

end function func_aux_2D_prima

function HH_isoclina_V_2D_prima(V) result(n_func)

     real(kind=dp), intent(in) :: V
     real(kind=dp)             :: n_func

     integer :: iostat_raiz

     V_aux = V

     call biseccion(func_aux_2D_prima,0.0_dp,1.0_dp,n_func,iostat_raiz)

end function HH_isoclina_V_2D_prima



! --- Isoclina cero n: ---

function HH_isoclina_n_2D(V) result(n)

     real(kind=dp), intent(in) :: V
     real(kind=dp)             :: n

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)
     n = n_inf_2D(y)

end function HH_isoclina_n_2D

function HH_isoclina_n_2D_prima(V) result(n)

     real(kind=dp), intent(in) :: V
     real(kind=dp)             :: n

     real(kind=dp), dimension(2) :: y

     y = (/V, null_float/)
     n = d_n_inf_2D(y)

end function HH_isoclina_n_2D_prima

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

end module m_modelo_HH_2D

