module m_tabla

! ::: MÓDULO: m_tabla ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!      Este módulo auxiliar contiene una subrutina de control de errores para la subrutina de búsqueda de raíces "biseccion".
!
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

public :: tabla_errores

integer, parameter, public :: RAIZ_OK = 0, RAIZ_IMAX = 1, RAIZ_SIGNO = 2

contains

subroutine tabla_errores(iostat_raiz)

integer, intent(in) :: iostat_raiz

if (iostat_raiz == 0) then
     print *, "Raíz calculada correctamente"
else if (iostat_raiz == 1) then
     print *, "Número de iteraciones máximas superado"
else if (iostat_raiz == 2) then
     print *, "No hay cambio de signo, buscar un intervalo más adecuado"
else
     print *, "Error desconocido"
end if

end subroutine tabla_errores

end module m_tabla





module m_subrutinas

! ::: MÓDULO: m_subrutinas :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!      Este módulo recoge los algoritmos de cálculo numérico y las subrutinas de generación de las gráficas en GnuPlot.
!
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

use m_parametros
use m_tabla

public  :: matrix_quick_sort, biseccion, rk4, gauss_jordan, DFT, eival, &
           string_to_number, gnu_file
private :: balanc, elmhes, hqr, eigsrt, &
           interval_shift, marker

integer, parameter, private :: IMAX_PREDETERMINADO = 100

contains

! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: MÉTODOS NUMÉRICOS

recursive subroutine matrix_quick_sort(matrix,row,col,order,abs_value)

! ::    Algoritmo quicksort de ordenación de valores aplicado a la fila (columna) especificada de la matriz de entrada. La columna (fila) 
! :: perpendicular se mantiene adherida al correspondiente valor a ordenar durante todo el algoritmo, conservando la relación entre diferentes
! :: filas (columnas). Ejm. Ordenar fila 2 de la matriz dada: 
! :: 
! ::        1  2  3  4  5         3  1  2  5  4
! ::        2  3  1  5  4   -->   1  2  3  4  5
! ::        1  2  3  4  5         3  1  2  5  4
! :: 
! :: Argumentos:
! ::    - matrix: Matriz de entrada. Es sustituida por la matriz ordenada según la fila (o columna) especificada.
! ::    - row, col (Variables enteras >= 0): Especifican la fila/columna a ordenar. Para ordenar la fila i (columna j) se debe asignar
! ::                                         row = i (col = j), mientras que la variable restante deberá tener valor nulo; 
! ::                                         col = 0 (row = 0).
! ::    - order (Opcional, variable lógica): .false. -> Ordena los valores de menor a mayor (valor predeterminado).
! ::                                         .true.  -> Ordena los valores de mayor a menor.
! ::    - abs_value (Opcional, variable lógica): .false. -> Ordena según valores originales (valor predeterminado).
! ::                                             .true.  -> Ordena según valores absolutos.

     real(kind=dp), dimension(:,:), intent(inout) :: matrix
     integer, intent(in)                          :: row, col
     logical, intent(in), optional                :: order, abs_value

     real(kind=dp), dimension(:), allocatable     :: pivot_value, temp
     integer                                      :: pivot_index, num_rows, num_cols, IOSTAT, i
     logical                                      :: qs_order, qs_abs, sorting_cond

     num_rows = size(matrix(:,1))
     num_cols = size(matrix(1,:))

     if (row == 0 .and. col == 0) then
          print *
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR MATRIX_QUICK_SORT: Fila/columna a ordenar no especificada    *"
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *
          return
     else if (row /= 0 .and. col /= 0) then
          print *
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR MATRIX_QUICK_SORT: Un único elemento de matriz especificado    *"
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *
          return
     end if

     if (col == 0) then
          if (row < 1 .or. row > num_rows) then
               print *
               print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
               print *, "*  ERROR MATRIX_QUICK_SORT: Número de fila especificado fuera de la matriz    *"
               print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
               print *
               return
          end if

          IOSTAT = 1 ! Algoritmo quicksort se va a aplicar a una fila
     else
          if (col < 1 .or. col > num_cols) then
               print *
               print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
               print *, "*  ERROR MATRIX_QUICK_SORT: Número de columna especificado fuera de la matriz     *"
               print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
               print *
               return
          end if

          IOSTAT = 2 ! Algoritmo quicksort se va a aplicar a una columna
     end if

     if (present(order)) then
          qs_order = order
     else
          qs_order = .false.
     end if

     if (present(abs_value)) then
          qs_abs = abs_value
     else
          qs_abs = .false.
     end if

     select case(IOSTAT)

     case(1) ! Algoritmo a una fila

          allocate( pivot_value(num_rows), temp(num_rows) )

          if (num_cols <= 1) then ! Array ya ordenado
               return
          end if

          pivot_value = matrix(:,num_cols)

          pivot_index = 1
          do i = 1, num_cols - 1 ! i marca la columna
               if (qs_order) then
                    if (qs_abs) then
                         sorting_cond = abs(matrix(row,i)) >= abs(pivot_value(row))
                    else
                         sorting_cond = matrix(row,i) >= pivot_value(row)
                    end if
               else
                    if (qs_abs) then
                         sorting_cond = abs(matrix(row,i)) <= abs(pivot_value(row))
                    else
                         sorting_cond = matrix(row,i) <= pivot_value(row)
                    end if
               end if

               if (sorting_cond) then
                    temp = matrix(:,i)
                    matrix(:,i) = matrix(:,pivot_index)
                    matrix(:,pivot_index) = temp
                    pivot_index = pivot_index + 1
               end if
          end do

          matrix(:,pivot_index+1:) = matrix(:,pivot_index:num_cols-1)
          matrix(:,pivot_index) = pivot_value

          call matrix_quick_sort(matrix(:,:pivot_index-1),row,col,order,abs_value)
          call matrix_quick_sort(matrix(:,pivot_index+1:),row,col,order,abs_value)

     case(2) ! Algoritmo a una columna

          allocate( pivot_value(num_cols), temp(num_cols) )

          if (num_rows <= 1) then ! Array ya ordenado
               return
          end if

          pivot_value = matrix(num_rows,:)

          pivot_index = 1
          do i = 1, num_rows - 1 ! i marca la fila
               if (qs_order) then
                    if (qs_abs) then
                         sorting_cond = abs(matrix(i,col)) >= abs(pivot_value(col))
                    else
                         sorting_cond = matrix(i,col) >= pivot_value(col)
                    end if
               else
                    if (qs_abs) then
                         sorting_cond = abs(matrix(i,col)) <= abs(pivot_value(col))
                    else
                         sorting_cond = matrix(i,col) <= pivot_value(col)
                    end if
               end if

               if (sorting_cond) then
                    temp = matrix(i,:)
                    matrix(i,:) = matrix(pivot_index,:)
                    matrix(pivot_index,:) = temp
                    pivot_index = pivot_index + 1
               end if
          end do

          matrix(pivot_index+1:,:) = matrix(pivot_index:num_rows-1,:)
          matrix(pivot_index,:) = pivot_value

          call matrix_quick_sort(matrix(:pivot_index-1,:),row,col,order,abs_value)
          call matrix_quick_sort(matrix(pivot_index+1:,:),row,col,order,abs_value)

     end select

     deallocate(pivot_value, temp)

end subroutine matrix_quick_sort



subroutine biseccion(funcion, lim_inf, lim_sup, raiz, iostat_raiz, precision, imax)

! ::    Algoritmo de búsqueda de raíces basado en el método de la bisección.
! :: 
! :: Argumentos:
! ::    - funcion: Debe indicar el nombre de la función a analizar. La función debe ser de variable real.
! ::    - lim_inf, lim_sup (Variables reales): Primera aproximación del intervalo en el que se encuentra la raiz.
! ::    - raiz (Variable real): Devuelve el valor final de la raiz.
! ::    - iostat_raiz (Variable entera): Variable para el control de errores (ver subrutina auxiliar "tabla_errores").
! ::    - precision (Opcional, variable real): Especifica la precisión a usar (valor predeterminado: PRECISION_DP).
! ::    - imax (Opcional, variable entera): Especifica el número máximo de iteraciones (valor predeterminado: IMAX_PREDETERMINADO).

     interface
          function funcion(x) result(f_x)
               use m_parametros
               real(kind=dp), intent(in) :: x
               real(kind=dp)             :: f_x
          end function funcion
     end interface
     real(kind=dp), intent(in)           :: lim_inf, lim_sup
     real(kind=dp), intent(out)          :: raiz
     integer, intent(out)                :: iostat_raiz
     real(kind=dp), intent(in), optional :: precision
     integer, intent(in), optional       :: imax

     integer       :: i, imaxx
     real(kind=dp) :: prec, p0, p1

     p0 = lim_inf
     p1 = lim_sup

     if (present(precision)) then
          prec = precision
     else
          prec = PRECISION_DP
     end if

     if (present(imax)) then
          imaxx = imax
     else
          imaxx = IMAX_PREDETERMINADO
     end if

     if (funcion(p0) * funcion(p1) > 0.0_dp) then
          iostat_raiz = RAIZ_SIGNO
          return
     end if

     do i = 1, imaxx
          raiz = (p0 + p1) / 2.0_dp

          if (abs(raiz - p0) <= prec) then
               iostat_raiz = RAIZ_OK
               return
          end if
          if (funcion(raiz) * funcion(p0) > 0.0_dp) then
               p0 = raiz
          else
               p1 = raiz
          end if
     end do

     iostat_raiz = RAIZ_IMAX

end subroutine biseccion



subroutine rk4(x,y,f,h)

! ::    Algoritmo Runge-Kutta de 4º orden de resolución de EDO.
! :: 
! :: Argumentos:
! ::    - x (Variable real): Variable independiente del sistema.
! ::    - y (Vector de componentes reales): Vector con n-1 componentes (n es el orden de la EDO) que recogen las condiciones iniciales.
! ::    - f: Nombre de la función a integrar, con argumentos x e y.
! ::    - h (Variable real): Magnitud del paso para la variable independiente.

     real(kind=dp), intent(inout)                :: x
     real(kind=dp), dimension(:), intent(inout)  :: y
     real(kind=dp), intent(in)                   :: h
     interface
     function f(x,y) result(yprima)
          use m_parametros
          real(kind=dp), intent(in)               :: x
          real(kind=dp), dimension(:), intent(in) :: y
          real(kind=dp), dimension(size(y))       :: yprima
     end function f
     end interface

     real(kind=dp), dimension(size(y)) :: k1, k2, k3, k4

     k1 = h * f(x,y)                    ! Euler
     k2 = h * f(x+0.5_dp*h,y+0.5_dp*k1) ! Punto medio
     k3 = h * f(x+0.5_dp*h,y+0.5_dp*k2) ! Otro punto medio
     k4 = h * f(x+h,y+k3)               ! Extremo final

     x = x + h
     y = y + (k1 + 2.0_dp*k2 + 2.0_dp*k3 +k4) / 6.0_dp ! Salto definitivo

end subroutine rk4



subroutine gauss_jordan(matrix_A,matrix_B)

! ::    Algoritmo de Gauss-Jordan para la resolución de sistemas de ecuaciones. La subrutina devuelve el resultado en el vector matrix_B.
! :: 
! :: Argumentos:
! ::    - matrix_A (Matriz de componentes reales): Matriz de coeficientes. 
! ::    - matrix_B (Vector de componentes reales): Vector de términos independientes. La solución se devuelve en este vector por lo que 
! ::                                               los valores iniciales se pierden tras la ejecución de la subrutina.

     real(kind=dp), dimension(:,:), intent(inout) :: matrix_A ! Matriz de coeficientes
     real(kind=dp), dimension(:), intent(inout)   :: matrix_B ! Términos independientes

     real(kind=dp), dimension(:,:), allocatable   :: augmented_matrix
     integer                                      :: num_rows, num_cols, i, j, k
     logical                                      :: all_cero

     num_rows = size(matrix_A(:,1))
     num_cols = size(matrix_A(1,:))

     allocate(augmented_matrix(num_rows,num_cols+1))

     if (num_rows /= size(matrix_B)) then
          deallocate(augmented_matrix)
          print *
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR GAUSS_JORDAN: Dimensiones de matrices incompatibles    *"
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *
          stop
     else if (num_rows /= num_cols) then
          deallocate(augmented_matrix)
          print *
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR GAUSS_JORDAN: Matriz de coeficientes no cuadrada     *"
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *
          stop
     end if

     augmented_matrix(:num_rows,:num_cols) = matrix_A
     augmented_matrix(:,num_cols+1) = matrix_B

     do j = 1, num_cols
          call matrix_quick_sort(augmented_matrix(j:,:),0,j,.true.,.true.) ! Selección de mayor elemento como pivote

          augmented_matrix(j,:) = augmented_matrix(j,:) / augmented_matrix(j,j) ! Pivote = 1
          do i = 1, num_rows
               if (i == j) then
                    cycle
               end if
               augmented_matrix(i,:) = augmented_matrix(i,:) - augmented_matrix(i,j) * augmented_matrix(j,:) ! Resto de elementos de la columna = 0
          end do
     end do

     matrix_B = augmented_matrix(:,num_cols+1)

     deallocate(augmented_matrix)

end subroutine gauss_jordan



function DFT(array) result(array_star)

! ::    Transformada de Fourier discreta. La función devuelve el resultado de la transformada en forma de vector complejo de una dimensión.
! :: 
! :: Argumentos:
! ::    - array (Vector de una dimensión, valores reales): Vector con los puntos a transformar.
! ::
! :: NOTA: Por afinidad con la teoría de las transformadas de Fourier, se adopta la numeración de vectores en Python (primer elemento marcado
! :: con el índice 0).

     real(kind=dp), dimension(:), intent(in)      :: array
     complex(kind=dp), dimension(0:size(array)-1) :: array_star

     complex(kind=dp), dimension(0:size(array)-1,0:size(array)-1) :: V_matrix ! Matriz de Vandermonde
     complex(kind=dp)                                             :: w
     integer                                                      :: j, k, n ! n: Número de puntos de la muestra

     real(kind=dp)    :: pi
     complex(kind=dp) :: i

     pi = acos(-1.0_dp)
     i = (0.0_dp, 1.0_dp)
     n = size(array)

     ! Configuración de la matriz de Vandermonde
     w = exp( (i * 2.0_dp * pi) / n )
     do k = 0, n - 1 ! Rows
          do j = 0, n - 1 ! Cols
               V_matrix(k,j) = w ** (k * j)
          end do
     end do

     array_star = matmul(V_matrix,array)

end function DFT



! --- Valores propios: ---

subroutine eival (a,wr,wi)

! ::    Algoritmo de cálculo de los valores propios para una matriz real general. Únicamente devuelve los valores propios de la matriz.
! :: 
! :: Argumentos:
! ::    - a (Matriz de componentes reales): Matriz de entrada. Tras la ejecución de la subrutina es destruida.
! ::    - wr (Vector de componentes reales): Partes reales de los valores propios ordenados de menor a mayor según sus módulos.
! ::    - wi (Vector de componentes reales): Partes imaginarias de los valores propios ordenados de menor a mayor según sus módulos.

real(kind=dp), intent(inout), dimension(:,:) :: a
real(kind=dp), intent(out), dimension(:)     :: wr,wi
integer :: n,i

n = size(a,dim=2)

if (size(a,dim=1) /= n) then

    stop "A is not square"

end if

if (size(wi) < n) then

        stop "wi is not big enough in eival"

end if

if (size(wr) < n) then

        stop "wr is not big enough in eival"

end if


! Matrix balancing

call balanc(a)

! Hessenberg form of the matrix

call elmhes(a)

! Eigenvalues

call hqr(a,wr,wi)

! Now we can order the eigenvalues using eigsrt(D,V). 
! D are the eigenvalue moduli. In V the first and the second
! column are the real and imaginary parts of the eigen values.
! Dirty. V is A. Also wr but it and wi are finally recovered in 
! the desired order. The matrix A is completely destroyed.

a      = 0.0_dp
a(1,:) = wr
a(2,:) = wi

do i=1,n

    wr(i) = sqrt(wr(i)*wr(i) + wi(i)*wi(i))

end do

call eigsrt(wr,a)

wr = a(1,:)
wi = a(2,:)

end subroutine eival

subroutine balanc(a)

real(kind=dp), intent(inout), dimension(:,:) :: a
real(kind=dp), parameter                     :: RADIX=2.0_dp,SQRDX=RADIX**2
integer                                      :: i,j,last,n
real(kind=dp)                                :: c,f,g,r,s

n = size(a,dim=2)
if (size(a,dim=1) /= n) then
    stop "A is not square"
end if

do
    last=1
    do i=1,n

        c=0.0_dp
        r=0.0_dp

        do j=1,n

            if(j /= i)then

                c=c+abs(a(j,i))
                r=r+abs(a(i,j))

            end if

        end do

        if(c /= 0.0_dp .and. r /= 0.0_dp)then

            g=r/RADIX
            f=1.0_dp
            s=c+r

            do

                if(c < g)then

                    f=f*RADIX
                    c=c*SQRDX

                else

                    exit

                end if

            end do

            g=r*RADIX

            do

                if(c > g)then

                    f=f/RADIX
                    c=c/SQRDX
          
                else
              
                    exit

                end if

            end do

                if((c+r)/f < 0.95_dp*s)then

                    last=0
                    g=1.0_dp/f

                do j=1,n

                    a(i,j)=a(i,j)*g
              
                end do
              
                do j=1,n
                  
                    a(j,i)=a(j,i)*f

                end do

            end if

        end if

    end do
      
    if(last /= 0) then
      
        exit

    end if

end do

end subroutine balanc
!  (C) Copr. 1986-92 Numerical Recipes Software Y].

subroutine elmhes(a)

real(kind=dp), intent(inout), dimension(:,:) :: a
integer                                      :: i,j,m,n
real                                         :: x,y

n = size(a,dim=2)

if (size(a,dim=1) /= n) then

    stop "A is not square"

end if

do m=2,n-1
        
    x=0.0_dp
    i=m

    do j=m,n

        if(abs(a(j,m-1)) > abs(x))then

            x=a(j,m-1)
            i=j

        end if

    end do

    if(i /= m)then

        do j=m-1,n

            y=a(i,j)
            a(i,j)=a(m,j)
            a(m,j)=y

        end do

        do j=1,n

            y=a(j,i)
            a(j,i)=a(j,m)
            a(j,m)=y

        end do

    end if

    if(x /= 0.0_dp)then

        do i=m+1,n

            y=a(i,m-1)

            if(y /= 0.0_dp)then

                y=y/x
                a(i,m-1)=y

                do j=m,n

                    a(i,j)=a(i,j)-y*a(m,j)

                end do

                do j=1,n

                    a(j,m)=a(j,m)+y*a(j,i)

                end do

            end if

        end do

    end if

end do

end subroutine elmhes
!  (C) Copr. 1986-92 Numerical Recipes Software Y].

subroutine hqr(a,wr,wi)


real(kind=dp), intent(inout), dimension(:,:) :: a
real(kind=dp), intent(out), dimension(:)     :: wi,wr
integer                                      :: i,its,j,k,l,m,nn,n
real(kind=dp)                                :: anorm,p,q,r,s,t,u,v,w,x,y,z
logical                                      :: premature_exit

n = size(a,dim=2)

if (size(a,dim=1) /= n) then

    stop "A is not square"

end if

if (size(wi) < n) then

        stop "wi is not big enough in hqr"

end if

if (size(wr) < n) then

        stop "wr is not big enough in hqr"

end if

anorm=abs(a(1,1))

do i=2,n

    do j=i-1,n

        anorm=anorm+abs(a(i,j))

    end do

end do

nn=n
t=0.0_dp

do_1: do

    if(nn >= 1)then

        its=0

do_2:   do

            premature_exit = .false. 

do_l:       do l=nn,2,-1

                s=abs(a(l-1,l-1))+abs(a(l,l))

                if(s == 0.0_dp) then

                    s=anorm

                end if

                if(abs(a(l,l-1))+s == s)then

                    premature_exit = .true.
                    exit do_l

                end if

            end do do_l

            if (.not. premature_exit) then

                l=1

            end if

            x=a(nn,nn)

            if(l == nn)then

                wr(nn)=x+t
                wi(nn)=0.0_dp
                nn=nn-1

            else
   
                y=a(nn-1,nn-1)
                w=a(nn,nn-1)*a(nn-1,nn)

                if(l == nn-1)then
           
                    p=0.5_dp*(y-x)
                    q=p**2.0_dp+w
                    z=sqrt(abs(q))
                    x=x+t
            
                    if(q >= 0.0_dp)then
                
                        z=p+sign(z,p)
                        wr(nn)=x+z
                        wr(nn-1)=wr(nn)
 
                        if(z /= 0.0_dp)then
     
                            wr(nn)=x-w/z

                        end if
        
                        wi(nn)=0.0_dp
                        wi(nn-1)=0.0_dp
 
                    else
               
                        wr(nn)=x+p
                        wr(nn-1)=wr(nn)
                        wi(nn)=z
                        wi(nn-1)=-z
            
                    end if
            
                    nn=nn-2
          
                else
        
                    if(its == 30)then
  
                        stop "too many iterations in hqr"
         
                    end if

                    if(its == 10 .or. its == 20)then
      
                        t=t+x
              
                        do i=1,nn
                
                            a(i,i)=a(i,i)-x
              
                        end do

                        s=abs(a(nn,nn-1))+abs(a(nn-1,nn-2))
                        x=0.75_dp*s
                        y=x
                        w=-0.4375_dp*s**2.0_dp

                    end if
      
                    its=its+1
            
do_m:               do m=nn-2,l,-1
              
                        z=a(m,m)
                        r=x-z
                        s=y-z
                        p=(r*s-w)/a(m+1,m)+a(m,m+1)
                        q=a(m+1,m+1)-z-r-s
                        r=a(m+2,m+1)
                        s=abs(p)+abs(q)+abs(r)
                        p=p/s
                        q=q/s
                        r=r/s
             
                        if(m == l) then
                  
                            exit do_m

                        end if
          
                        u=abs(a(m,m-1))*(abs(q)+abs(r))
                        v=abs(p)*(abs(a(m-1,m-1))+abs(z)+abs(a(m+1,m+1)))
              
                        if(u+v == v)then
                  
                            exit do_m
             
                        end if

                    end do do_m

                    do i=m+2,nn
              
                        a(i,i-2)=0.0_dp
                    
                        if (i /= m+2) then
                  
                            a(i,i-3)=0.0_dp
              
                        end if
            
                    end do
            
                    do k=m,nn-1
              
                        if(k /= m)then
                
                            p=a(k,k-1)
                            q=a(k+1,k-1)
                            r=0.0_dp
              
                            if(k /= nn-1)then
                    
                                r=a(k+2,k-1)
                
                            end if
               
                            x=abs(p)+abs(q)+abs(r)
               
                            if(x /= 0.0_dp)then
                  
                                p=p/x
                                q=q/x
                                r=r/x
                
                            end if
              
                        end if

                        s=sign(sqrt(p**2.0_dp+q**2.0_dp+r**2.0_dp),p)
        
                        if(s /= 0.0_dp)then
                
                            if(k == m)then
                  
                                if(l /= m)then
                      
                                    a(k,k-1)=-a(k,k-1)
                  
                                end if
               
                            else
                  
                                a(k,k-1)=-s*x
                
                            end if
                
                            p=p+s
                            x=p/s
                            y=q/s
                            z=r/s
                            q=q/p
                            r=r/p
                
                            do j=k,nn
                  
                                p=a(k,j)+q*a(k+1,j)
                  
                                if(k /= nn-1)then
                    
                                    p=p+r*a(k+2,j)
                                    a(k+2,j)=a(k+2,j)-p*z
                  
                                end if
                
                                a(k+1,j)=a(k+1,j)-p*y
                                a(k,j)=a(k,j)-p*x
                
                            end do
                
                            do i=l,min(nn,k+3)
                   
                                p=x*a(i,k)+y*a(i,k+1)
                  
                                if(k /= nn-1)then
                 
                                    p=p+z*a(i,k+2)
                                    a(i,k+2)=a(i,k+2)-p*r
                  
                                end if
                  
                                a(i,k+1)=a(i,k+1)-p*q
                                a(i,k)=a(i,k)-p
                
                            end do
                        
                        end if

                    end do

                    cycle do_2

                end if

            end if

            exit do_2

            end do do_2
      
    else

            exit do_1
   
    end if

end do do_1

end subroutine hqr
!  (C) Copr. 1986-92 Numerical Recipes Software Y].

subroutine eigsrt(D,V)

!INTEGER, intent(in) ::  N,NP

real(kind=dp), intent(inout), dimension(:)             ::  D
real(kind=dp), intent(inout), dimension(:,:), optional ::  V

real(kind=dp)    :: P
integer :: I,J,K,n

n = size(d)

DO I = 1,N - 1
    K = I
    P = D(I)
    DO J = I + 1,N
        IF (D(J) <= P) THEN
            K = J
            P = D(J)
        END IF
    enddo
    IF (K /= I) THEN
        D(K) = D(I)
        D(I) = P
        if (present(v)) then
            DO J = 1,N
                P = V(J,I)
                V(J,I) = V(J,K)
                V(J,K) = P
            enddo
        end if
    END IF
ENDDO

end subroutine eigsrt

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::



! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::: GENERACIÓN DE SCRIPTS GNUPLOT

subroutine string_to_number(str,num)

     character(len=*), intent(inout)          :: str
     real(kind=dp), dimension(:), intent(out) :: num ! dim = 3 (número max. de puntos de eq.)

     character(len=30) :: str_temp
     integer :: i, j

     str_temp = ""
     j = 1

     if (str(1:5) == "array") then
          do i = 1, len(str)
               if (str(i:i) == "=") then
                    str = str(i+2:)
               end if
          end do
          do i = 2, len(str)
               if (str(i:i) == ",") then
                    if (len(trim(str_temp)) == 0) then
                         str_temp = "1000.0"
                    end if
                    read(str_temp,fmt=*) num(j)
                    j = j + 1
                    str_temp = ""
               else if (str(i:i) == "]") then
                    if (len(trim(str_temp)) == 0) then
                         str_temp = "1000.0"
                    end if
                    read(str_temp,fmt=*) num(j)
                    exit
               else
                    str_temp = trim(str_temp) // str(i:i)
               end if
          end do
     else if (str(1:2) == "m_") then
          read(str(7:),fmt=*) num(j)
          num(j + 1) = 1000.0_dp
          num(j + 2) = 1000.0_dp
     else
          print *
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR STRING_TO_NUMBER: Cadena de caracteres no reconocida     *"
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *
          stop
     end if

end subroutine string_to_number



subroutine gnu_file(IOSTAT,b,num_ptos_eq,mouse_xpos,mouse_ypos,info_ptos_eq)

! ::    Subrutina de escritura de scripts de GnuPlot para la generación de gráficas.
! :: 
! :: Resumen de índices IOSTAT:
! ::          IOSTAT 3: Integración del modelo 4D de Hodgkin-Huxley
! ::          IOSTAT 4: Funciones de activación/inactivación y constantes de tiempo (4D)
! ::          IOSTAT 5: Integración del modelo 2D de Hodgkin-Huxley
! ::          IOSTAT 6: Función de activación y constante de tiempo del K+ (2D)
! ::          IOSTAT 10: Cálculo de puntos de equilibrio + estabilidad
! ::          IOSTAT 15: Cálculo de los valores de bifurcación de E_K (Bif. Hopf)
! ::          IOSTAT 16: Análisis numérico y gráficas (Bif. Hopf)
! ::          IOSTAT 17: Cálculo de los valores de bifurcación de E_K (Bif. SN)
! ::          IOSTAT 18: Análisis numérico y gráficas (Bif. SN)
! ::          IOSTAT 13: Bifurcación SN de ciclos límite
! ::          IOSTAT 19: Recogida de puntos de equilibrio (Esp. fases)
! ::          IOSTAT 20: Generación de gráfica (Esp. fases)
! ::          IOSTAT 23: DFT

     integer, intent(in)                               :: IOSTAT
     real(kind=dp), intent(in), optional               :: b
     integer, intent(in), optional                     :: num_ptos_eq
     real(kind=dp), dimension(:), intent(in), optional :: mouse_xpos, mouse_ypos
     integer, dimension(:), intent(in), optional       :: info_ptos_eq

     real(kind=dp), dimension(2)                       :: point_coord
     character(len=512)                                :: str_temp1, str_temp2, str_temp3
     character(len=1024)                               :: gnu_string
     integer                                           :: info_graph, i

select case(IOSTAT)

! --- case(1): Modelo de Hodgkin-Huxley original ---
case(3) ! Integración del modelo 4D de Hodgkin-Huxley

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'modelo_HH_4D.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key off"
     write(unit=1,fmt=*) "set xlabel 't (ms)'"
     write(unit=1,fmt=*) "set ylabel 'V_{mem} (mV)' offset -1,0,0"
     write(unit=1,fmt=*) "plot [0:200] [-80:80] 'modelo_HH_4D.txt' using 1:2 w l lt -1"

     close(unit=1)

case(4) ! Funciones de activación/inactivación y constantes de tiempo

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'func_act_inact_4D.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key spacing 1.5"
     write(unit=1,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=1,fmt=*) "set ylabel 'Funciones de activación / inactivación'"
     write(unit=1,fmt=*) "plot [-100:50] [0:1] 'modelo_HH_4D_caracteristicas.txt' index 0 using 1:2 w l lt 2 &
                         &title 'm_{/Symbol \245}(V)', '' index 0 using 1:3 w l lt 4 title 'h_{/Symbol \245}(V)', &
                         &'' index 0 using 1:4 w l lt 6 title 'n_{/Symbol \245}(V)'"

     write(unit=1,fmt=*) "set output 'ctes_tiempo_4D.pdf'"
     write(unit=1,fmt=*) "set ylabel 'Constantes de tiempo (ms)'"
     write(unit=1,fmt=*) "plot [-100:80] [0:9] 'modelo_HH_4D_caracteristicas.txt' index 1 using 1:2 w l lt 2 &
                         &title '{/Symbol t}_m(V)', '' index 1 using 1:3 w l lt 4 title '{/Symbol t}_h(V)', &
                         &'' index 1 using 1:4 w l lt 6 title '{/Symbol t}_n(V)'"

     close(unit=1)

! --- case(2): Modelo de Hodgkin-Huxley reducido ---
case(5) ! Integración del modelo 2D de Hodgkin-Huxley

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'modelo_HH_2D.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key off"
     write(unit=1,fmt=*) "set xlabel 't (ms)'"
     write(unit=1,fmt=*) "set ylabel 'V_{mem} (mV)' offset -1,0,0"
     write(unit=1,fmt=*) "plot [0:200] [-80:80] 'modelo_HH_2D.txt' using 1:2 w l lt -1"

     close(unit=1)

case(6) ! Función de activación y constante de tiempo del K+

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'func_act_inact_2D.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=1,fmt=*) "set ylabel 'Función activación K^+'"
     write(unit=1,fmt=*) "plot [-100:50] [0:1] 'modelo_HH_2D_caracteristicas.txt' index 0 w l lt -1 title 'n_{/Symbol \245}(V)'"

     write(unit=1,fmt=*) "set output 'cte_tiempo_2D.pdf'"
     write(unit=1,fmt=*) "set ylabel 'Constante de tiempo K^+ (ms)'"
     write(unit=1,fmt=*) "plot [-100:80] [0:9] 'modelo_HH_2D_caracteristicas.txt' index 1 w l lt -1 title '{/Symbol t}_n(V)'"

     close(unit=1)

case(10) ! Cálculo de puntos de equilibrio + estabilidad

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term wxt"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key bottom left at -85,0.045"
     write(unit=1,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=1,fmt=*) "set ylabel 'n (Variable de activación de K^+)'"
     write(unit=1,fmt=*) "plot [-80:60] [0:1] 'isoclinas_cero_temp.txt' index 0 w l lt -1 title 'Isoclina cero {/Times:Italic n}', &
                         &'isoclinas_cero_temp.txt' index 1 w l lt 1 title 'Isoclina cero {/Times:Italic V_{mem}}'"
     write(unit=1,fmt=*) "pause mouse"
     write(unit=1,fmt=*) "m_x = MOUSE_X"
     write(unit=1,fmt=*) "m_y = MOUSE_Y"
     write(unit=1,fmt=*) "save variables 'var_temp.txt'"

     close(unit=1)

! --- case(11): Bifurcación de Hopf ---
case(15) ! Cálculo de los valores de bifurcación de E_K

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term wxt"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key off"
     write(unit=1,fmt=*) "set multiplot layout 1,2"
     write(unit=1,fmt=*) "set xlabel 'E_K (mV)'"
     write(unit=1,fmt=*) "set ylabel 'Tr  J(V*,n*)'"
     write(unit=1,fmt=*) "set xzeroaxis"
     write(unit=1,fmt=*) "set title 'Bif. Hopf I'"
     write(unit=1,fmt=*) "plot [-77:-71] [-2:5] 'trJ_temp.txt' w l lt -1"
     write(unit=1,fmt=*) "unset title"
     write(unit=1,fmt=*) "set title 'Bif. Hopf II'"
     write(unit=1,fmt=*) "plot [-48:-42] [-2:5] 'trJ_temp.txt'  w l lt -1"
     write(unit=1,fmt=*) "unset title"
     write(unit=1,fmt=*) "unset multiplot"
     write(unit=1,fmt=*) "pause mouse"
     write(unit=1,fmt=*) "m_x = MOUSE_X"
     write(unit=1,fmt=*) "m_y = MOUSE_Y"
     write(unit=1,fmt=*) "save variables 'var_temp.txt'"

     close(unit=1)

case(16) ! Análisis numérico y gráficas

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'tr_J.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key off"
     write(unit=1,fmt=*) "set xlabel 'E_K (mV)'"
     write(unit=1,fmt=*) "set ylabel 'Tr  J(V*,n*)'"
     write(unit=1,fmt=*) "set xzeroaxis"

     if (E_K < -60.0_dp) then
          info_graph = 3
          null_farray2 = marker_coordinates
          call marker(marker_coordinates,7,info_graph,gnu_string)
          write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
          write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
          write(unit=1,fmt=*) "plot [] [-2:5] 'bifurcacion_hopf.txt' index 0 w l lt -1"
          write(unit=1,fmt=*) "unset object"

          write(unit=1,fmt=*) "set output 'autovalores.pdf'"
          write(unit=1,fmt=*) "set size ratio 1"
          write(unit=1,fmt=*) "set key at graph 0.97,0.97"
          write(unit=1,fmt=*) "unset ylabel"
          write(unit=1,fmt=*) "set xzeroaxis"

          info_graph = 6
          marker_coordinates = null_farray2
          marker_coordinates(2) = 0.0_dp
          call marker(marker_coordinates,7,info_graph,gnu_string)
          write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
          write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))

          write(unit=1,fmt=*) "plot [-74.492:-73.492] [-0.6:0.6] 'bifurcacion_hopf.txt' &
                              &index 1 using 1:2 w l lt 7 title 'c_{1,2} (E_K)', '' index 1 using 1:3 w l lt 6 &
                              &title 'ω_{1,2} (E_K)', '' index 1 using 1:5 w l lt 6 title ''"
     else
          info_graph = 4
          null_farray2 = marker_coordinates
          call marker(marker_coordinates,7,info_graph,gnu_string)
          write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
          write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
          write(unit=1,fmt=*) "plot [] [-2:5] 'bifurcacion_hopf.txt' index 0 w l lt -1"
          write(unit=1,fmt=*) "unset object"

          write(unit=1,fmt=*) "set output 'autovalores.pdf'"
          write(unit=1,fmt=*) "set size ratio 1"
          write(unit=1,fmt=*) "set key at graph 0.97,0.97"
          write(unit=1,fmt=*) "unset ylabel"
          write(unit=1,fmt=*) "set xzeroaxis"

          info_graph = 8
          marker_coordinates = null_farray2
          marker_coordinates(2) = 0.0_dp
          call marker(marker_coordinates,7,info_graph,gnu_string)
          write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
          write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))

          write(unit=1,fmt=*) "plot [-45.725:-44.725] [-6.5:6.5] 'bifurcacion_hopf.txt' &
                              &index 1 using 1:2 w l lt 7 title 'c_{1,2} (E_K)', '' index 1 using 1:3 w l lt 6 &
                              &title 'ω_{1,2} (E_K)', '' index 1 using 1:5 w l lt 6 title ''"
     end if

     close(unit=1)

! --- case(12): Bifurcación SN ---
case(17) ! Cálculo de los valores de bifurcación de E_K

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term wxt"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key off"
     write(unit=1,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=1,fmt=*) "set ylabel 'E_K = E_K(V_{mem}) (mV)'"
     write(unit=1,fmt=*) "plot [-80:0] [-100:-20] 'bifurcacion_SN.txt' using 1:2 w l lt -1"
     write(unit=1,fmt=*) "pause mouse"
     write(unit=1,fmt=*) "m_x = MOUSE_X"
     write(unit=1,fmt=*) "m_y = MOUSE_Y"
     write(unit=1,fmt=*) "save variables 'var_temp.txt'"

     close(unit=1)

case(18) ! Análisis numérico y gráficas

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'bifurcacion_SN.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key off"

     write(unit=1,fmt=*) "set multiplot layout 1,2"
     write(unit=1,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=1,fmt=*) "set ylabel 'E_K = E_K(V_{mem}) (mV)'"
     info_graph = 11
!marker_coordinates = (/-43.315884583631629_dp, -72.6390479574_dp/)
     null_farray2 = marker_coordinates
     call marker(marker_coordinates,7,info_graph,gnu_string)
     write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
     write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
!marker_coordinates = (/-55.297238171303448_dp, -68.6593233834_dp/)
!call marker(marker_coordinates,7,info_graph,gnu_string)
!write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
!write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
     write(unit=1,fmt=*) "plot [-80:0] [-100:-20] 'bifurcacion_SN.txt' using 1:2 w l lt -1"
     write(unit=1,fmt=*) "unset object"

     write(unit=1,fmt=*) "set ylabel 'dE_K(V_{mem}) / dV_{mem} (mV)'"
     write(unit=1,fmt=*) "set xzeroaxis"
     info_graph = 12
!marker_coordinates = (/-43.315884583631629_dp, 0.0_dp/)
     marker_coordinates = null_farray2
     marker_coordinates(2) = 0.0_dp
     call marker(marker_coordinates,7,info_graph,gnu_string)
     write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
     write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
!marker_coordinates = (/-55.297238171303448_dp, 0.0_dp/)
!call marker(marker_coordinates,7,info_graph,gnu_string)
!write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
!write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
     write(unit=1,fmt=*) "plot [-80:0] [-5:10] 'bifurcacion_SN.txt' using 1:3 w l lt -1"
     write(unit=1,fmt=*) "unset multiplot"

     close(unit=1)

case(13) ! Bifurcación SN de ciclos límite

     if (diagrama_bif_total) then
          info_graph = 2
     else
          info_graph = 1
     end if

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'bifurcacion_SN_ciclos_limite.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set key top right at graph 0.97,0.95"
     write(unit=1,fmt=*) "set xlabel 'E_K (mV)'"
     write(unit=1,fmt=*) "set ylabel 'Amplitud ciclos límite (n_{max} - n_{min})'"

     call marker(marker_coordinates,7,info_graph,gnu_string)
     write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
     write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))

     if (diagrama_bif_total) then
          write(unit=1,fmt=*) "plot [-80:-40] [0:1] 'bifurcacion_SN_ciclos_limite.txt' using 1:2 w l lt 4 lw 1.5 &
                              &title 'Ciclo límite estable', '' using 1:3 w l lt 2 lw 1.5 title 'Ciclo límite inestable'"
     else
          write(unit=1,fmt=*) "plot [-77:-73.5] [0:1] 'bifurcacion_SN_ciclos_limite.txt' using 1:2 w l lt 4 lw 1.5 &
                              &title 'Ciclo límite estable', '' using 1:3 w l lt 2 lw 1.5 title 'Ciclo límite inestable'"
     end if

     close(unit=1)

! --- case(14): Representación del espacio de fases ---
case(19) ! Recogida de puntos de equilibrio

     open(unit=2, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=2,fmt=*) "set term wxt"
     write(unit=2,fmt=*) "set mouse"
     write(unit=2,fmt=*) "set size ratio 1"
     write(unit=2,fmt=*) "set key bottom left at -85,0.045"
     write(unit=2,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=2,fmt=*) "set ylabel 'n (Variable de activación de K^+)'"
     write(unit=2,fmt=*) "array mouse_xpos[3]"
     write(unit=2,fmt=*) "array mouse_ypos[3]"
     write(unit=2,fmt=*) "print '   --- Comienzo de recogida de datos ---'"
     write(unit=2,fmt=*) "print ' '"
     write(unit=2,fmt=*) "print '      Esperando punto...'"
     write(unit=2,fmt=*) "status = 0"
     write(unit=2,fmt=*) "do for [i = 1:3] {"
     write(unit=2,fmt=*) "     plot [-80:60] [0:1] 'espacio_fases.txt' index 0 w l lt -1 &
                               &title 'Isoclina cero {/Times:Italic n}', 'espacio_fases.txt' index 1 w l lt 1 &
                               &title 'Isoclina cero {/Times:Italic V_{mem}}'"
     write(unit=2,fmt=*) "     pause mouse"
     write(unit=2,fmt=*) "     mouse_xpos[i] = MOUSE_X"
     write(unit=2,fmt=*) "     mouse_ypos[i] = MOUSE_Y"
     write(unit=2,fmt=*) "     print '      Punto registrado'"
     write(unit=2,fmt=*) "     if (i != 3) {"
     write(unit=2,fmt=*) "          pause mouse keypress"
     write(unit=2,fmt=*) "          if (MOUSE_KEY == 9) {"
     write(unit=2,fmt=*) "               print '      Esperando punto...'"
     write(unit=2,fmt=*) "          }"
     write(unit=2,fmt=*) "          else {"
     write(unit=2,fmt=*) "               if (MOUSE_KEY == 13) {"
     write(unit=2,fmt=*) "                    status = 1"
     write(unit=2,fmt=*) "                    break"
     write(unit=2,fmt=*) "               }"
     write(unit=2,fmt=*) "               else {"
     write(unit=2,fmt=*) "                    status = -1"
     write(unit=2,fmt=*) "                    break"
     write(unit=2,fmt=*) "               }"
     write(unit=2,fmt=*) "          }"
     write(unit=2,fmt=*) "     }"
     write(unit=2,fmt=*) "}"
     write(unit=2,fmt=*) "print ' '"
     write(unit=2,fmt=*) "if (status == -1) {"
     write(unit=2,fmt=*) "     print '   --- Tecla no soportada, continuando programa con puntos registrados... ---'"
     write(unit=2,fmt=*) "     save variables 'var_temp.txt'"
     write(unit=2,fmt=*) "}"
     write(unit=2,fmt=*) "else {"
     write(unit=2,fmt=*) "     if (status == 0) {"
     write(unit=2,fmt=*) "          print '   --- Número máximo de puntos alcanzado, fin de recogida de datos ---'"
     write(unit=2,fmt=*) "          save variables 'var_temp.txt'"
     write(unit=2,fmt=*) "     }"
     write(unit=2,fmt=*) "     else {"
     write(unit=2,fmt=*) "          print '   --- Fin de recogida de datos ---'"
     write(unit=2,fmt=*) "          save variables 'var_temp.txt'"
     write(unit=2,fmt=*) "     }"
     write(unit=2,fmt=*) "}"

     close(unit=2)

case(20) ! Generación de gráfica

     info_graph = 13

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'espacio_fases.pdf'"
     write(unit=1,fmt=*) "set size ratio 1"
     write(unit=1,fmt=*) "set xlabel 'V_{mem} (mV)'"
     write(unit=1,fmt=*) "set ylabel 'n (Variable de activación de K^+)'"

     ! --- Representación de puntos de equilibrio: ---
     do i = 1, num_ptos_eq
          point_coord = (/mouse_xpos(i), mouse_ypos(i)/)
          call marker(point_coord,info_ptos_eq(i),info_graph,gnu_string)
          if (info_ptos_eq(i) == 5 .or. info_ptos_eq(i) == 7) then
               write(unit=1,fmt=*) trim(gnu_string(:scan(gnu_string,"/")-1))
               write(unit=1,fmt=*) trim(gnu_string(scan(gnu_string,"/")+1:))
          else
               write(unit=1,fmt=*) trim(gnu_string)
          end if
     end do

     if (b >= rango(1) .and. b < rango(2)) then
          write(unit=1,fmt=*) "set key top right at graph 0.97,0.95 font ',9'"
          write(unit=1,fmt=*) "plot [-80:60] [0:1] 'espacio_fases.txt' index 0 w l lt -1 dt 4 title '', &
                              &'' index 1 w l lt -1 title '', &
                              &'' index 2 w l lt 4 lw 1.5 title 'Ciclo límite estable', &
                              &'' index 3 w l lt 2 lw 1.5 title 'Ciclo límite inestable'"
     else if (b >= rango(2) .and. b < rango(5)) then
          write(unit=1,fmt=*) "set key top right at graph 0.97,0.95 font ',9'"
          write(unit=1,fmt=*) "plot [-80:60] [0:1] 'espacio_fases.txt' index 0 w l lt -1 dt 4 title '', &
                              &'' index 1 w l lt -1 title '', &
                              &'' index 2 w l lt 4 lw 1.5 title 'Ciclo límite estable'"
     else
          write(unit=1,fmt=*) "set key off"
          write(unit=1,fmt=*) "plot [-80:60] [0:1] 'espacio_fases.txt' index 0 w l lt -1 dt 4, '' index 1 w l lt -1"
     end if

     close(unit=1)

case(23) ! DFT

     open(unit=1, file="gnu_temp.gnu", status="replace", action="write")

     write(unit=1,fmt=*) "set term pdf"
     write(unit=1,fmt=*) "set output 'DFT.pdf'"
     write(unit=1,fmt=*) "set key off"
     write(unit=1,fmt=*) "set multiplot layout 3,1"
     write(unit=1,fmt=*) "set lmargin 9"
     write(unit=1,fmt=*) "set rmargin "
     write(unit=1,fmt=*) "set xlabel 't (ms)'"
     write(unit=1,fmt=*) "set ylabel 'I(t) (μA/cm^2)'"
     write(unit=1,fmt=*) "set ytics -0.2,0.1,0.2"
     write(unit=1,fmt=*) "plot [0:1000] [-0.2:0.2] 'DFT.txt' index 1 w l lt -1"
     write(unit=1,fmt=*) "set ylabel 'ΔV_{mem} (mV)'"
     write(unit=1,fmt=*) "set ytics -1.5,0.5,1.5"
     write(unit=1,fmt=*) "plot [0:1000] [-1.5:1.5] 'DFT.txt' index 0 w l lt -1"
     write(unit=1,fmt=*) "set xlabel 'f (Hz)'"
     write(unit=1,fmt=*) "set ylabel 'Z(f) (MOhm)' offset 0.5,0,0"
     write(unit=1,fmt=*) "set ytics 0,0.02,0.04"
     write(unit=1,fmt=*) "plot [0:500] [0:0.04] 'DFT.txt' index 2 w l lt -1"
     write(unit=1,fmt=*) "unset multiplot"

     close(unit=1)

case default

     print *
     print *, "* * * * * * * * * * * * * * * * * * * * *"
     print *, "*  ERROR GNU_FILE: IOSTAT no válido     *"
     print *, "* * * * * * * * * * * * * * * * * * * * *"
     print *
     stop

end select

end subroutine gnu_file

function interval_shift(x_old,info_graph) result(x_new)

     real(kind=dp), intent(in)   :: x_old
     integer, intent(in)         :: info_graph
     real(kind=dp)               :: x_new

     real(kind=dp), dimension(2) :: old_interval, new_interval

     if (info_graph == 1) then ! SNCL parcial (OX)
          old_interval = (/-77.0_dp, -73.5_dp/)
     else if (info_graph == 2) then ! SNCL completa (OX)
          old_interval = (/-80.0_dp, -40.0_dp/)

     else if (info_graph == 3) then ! Traza Hopf I (OX)
          old_interval = (/-77.0_dp, -71.0_dp/)
     else if (info_graph == 4) then ! Traza Hopf II (OX)
          old_interval = (/-48.0_dp, -42.0_dp/)
     else if (info_graph == 5) then ! Traza Hopf (OY)
          old_interval = (/-2.0_dp, 5.0_dp/)
     else if (info_graph == 6) then ! VP Hopf I (OX)
          old_interval = (/-74.492_dp, -73.492_dp/)
     else if (info_graph == 7) then ! VP Hopf I (OY)
          old_interval = (/-0.6_dp, 0.6_dp/)
     else if (info_graph == 8) then ! VP Hopf II (OX)
          old_interval = (/-45.725_dp, -44.725_dp/)
     else if (info_graph == 9) then ! VP Hopf II (OY)
          old_interval = (/-6.5_dp, 6.5_dp/)

     else if (info_graph == 10) then ! SN (OX)
          old_interval = (/-80.0_dp, 0.0_dp/)
     else if (info_graph == 11) then ! SN 1 (OY)
          old_interval = (/-100.0_dp, -20.0_dp/)
     else if (info_graph == 12) then ! SN 2 (OY)
          old_interval = (/-5.0_dp, 10.0_dp/)

     else if (info_graph == 13) then ! Espacio de fases (OX)
          old_interval = (/-80.0_dp, 60.0_dp/)
     else
          print *
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR INTERVAL_SHIFT: info_graph no válido     *"
          print *, "* * * * * * * * * * * * * * * * * * * * * * * * * *"
          print *
          stop
     end if

     new_interval = (/0.0_dp, 1.0_dp/)

     x_new = new_interval(1) + ( (new_interval(2) - new_interval(1)) / (old_interval(2) - old_interval(1)) ) &
                             * (x_old - old_interval(1))

end function interval_shift

subroutine marker(point_coord,info_pto_eq,info_graph,gnu_string)

     real(kind=dp), dimension(:), intent(inout) :: point_coord
     integer, intent(in)                        :: info_pto_eq, info_graph
     character(len=*), intent(out)              :: gnu_string

     real(kind=dp)                              :: height, length, side, radius, apothem, theta
     real(kind=dp), dimension(2,2)              :: rotation_matrix
     real(kind=dp), dimension(2,4)              :: vertices
     character(len=512)                         :: str_temp1, str_temp2

     gnu_string = ""
     str_temp1 = ""
     str_temp2 = ""
     if (info_pto_eq == 1) then
          gnu_string = trim(gnu_string) // "set object rectangle front at "
          write(str_temp1,fmt=*) point_coord(1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) point_coord(2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " size graph 0.018,0.018 fillcolor rgb 'black' &
                                                            &fillstyle solid border lt -1"
     else if (info_pto_eq == 2) then
          gnu_string = trim(gnu_string) // "set object rectangle front at "
          write(str_temp1,fmt=*) point_coord(1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) point_coord(2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " size graph 0.018,0.018 fillcolor rgb 'white' &
                                                            &fillstyle solid border lt -1"
     else if (info_pto_eq == 3) then
          gnu_string = trim(gnu_string) // "set object circle front at "
          write(str_temp1,fmt=*) point_coord(1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) point_coord(2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " radius graph 0.01 fillcolor rgb 'black' &
                                                            &fillstyle solid border lt -1"
     else if (info_pto_eq == 4) then
          gnu_string = trim(gnu_string) // "set object circle front at "
          write(str_temp1,fmt=*) point_coord(1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) point_coord(2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " radius graph 0.01 fillcolor rgb 'white' &
                                                            &fillstyle solid border lt -1"
     else if (info_pto_eq == 5) then
          gnu_string = trim(gnu_string) // "set object circle front at "
          write(str_temp1,fmt=*) point_coord(1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) point_coord(2)
          str_temp2 = trim(gnu_string) // trim(str_temp1) // " radius graph 0.01 arc [90:-90] fillcolor rgb 'black' &
                                                            &fillstyle solid noborder"
          gnu_string = trim(gnu_string) // trim(str_temp1) // " radius graph 0.01 fillcolor rgb 'white' &
                                                            &fillstyle solid border lt -1"
          gnu_string = trim(gnu_string) // "/" // trim(str_temp2)
     else if (info_pto_eq == 6) then
          height = 0.025_dp

          side = height / sin( acos(-1.0_dp) / 3.0_dp )
          radius = side / sqrt(3.0_dp)
          apothem = height - radius

          if (info_graph == 3 .or. info_graph == 4) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),5)
          else if (info_graph == 6) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),7)
          else if (info_graph == 8) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),9)
          else if (info_graph == 11 .or. info_graph == 12) then
               point_coord(1) = interval_shift(point_coord(1),10)
               point_coord(2) = interval_shift(point_coord(2),info_graph)
          else
               point_coord(1) = interval_shift(point_coord(1),info_graph)
          end if

          vertices(:,1) = (/point_coord(1) - (side / 2.0_dp), point_coord(2) - apothem/)
          vertices(:,2) = (/point_coord(1) + (side / 2.0_dp), point_coord(2) - apothem/)
          vertices(:,3) = (/point_coord(1), point_coord(2) + radius/)

          gnu_string = ""
          gnu_string = trim(gnu_string) // "set object polygon front from graph "
          write(str_temp1,fmt=*) vertices(1,1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
          write(str_temp1,fmt=*) vertices(1,2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
          write(str_temp1,fmt=*) vertices(1,3)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,3)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
          write(str_temp1,fmt=*) vertices(1,1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,1)
          gnu_string = trim(gnu_string) // trim(str_temp1)
          gnu_string = trim(gnu_string) // " fillcolor rgb '#B860CD' fillstyle solid noborder"
     else if (info_pto_eq == 7) then
!          height = 0.01_dp
!          length = 0.035_dp
          height = 0.008_dp
          length = 0.030_dp

          if (info_graph == 3 .or. info_graph == 4) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),5)
          else if (info_graph == 6) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),7)
          else if (info_graph == 8) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),9)
          else if (info_graph == 11 .or. info_graph == 12) then
               point_coord(1) = interval_shift(point_coord(1),10)
               point_coord(2) = interval_shift(point_coord(2),info_graph)
          else
               point_coord(1) = interval_shift(point_coord(1),info_graph)
          end if

          do i = 0, 1
               vertices(:,1) = (/-(length / 2.0_dp),  (height / 2.0_dp)/)
               vertices(:,2) = (/ (length / 2.0_dp),  (height / 2.0_dp)/)
               vertices(:,3) = (/ (length / 2.0_dp), -(height / 2.0_dp)/)
               vertices(:,4) = (/-(length / 2.0_dp), -(height / 2.0_dp)/)

               theta = (-1.0_dp)**i * (acos(-1.0_dp) / 4.0_dp)
               rotation_matrix = reshape((/cos(theta),sin(theta),-sin(theta),cos(theta)/), (/2,2/))

               vertices(:,1) = matmul(rotation_matrix,vertices(:,1))
               vertices(:,2) = matmul(rotation_matrix,vertices(:,2))
               vertices(:,3) = matmul(rotation_matrix,vertices(:,3))
               vertices(:,4) = matmul(rotation_matrix,vertices(:,4))

               vertices(:,1) = (/vertices(1,1) + point_coord(1), vertices(2,1) + point_coord(2)/)
               vertices(:,2) = (/vertices(1,2) + point_coord(1), vertices(2,2) + point_coord(2)/)
               vertices(:,3) = (/vertices(1,3) + point_coord(1), vertices(2,3) + point_coord(2)/)
               vertices(:,4) = (/vertices(1,4) + point_coord(1), vertices(2,4) + point_coord(2)/)

               gnu_string = trim(gnu_string) // "set object polygon front from graph "
               write(str_temp1,fmt=*) vertices(1,1)
               gnu_string = trim(gnu_string) // trim(str_temp1) // ","
               write(str_temp1,fmt=*) vertices(2,1)
               gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
               write(str_temp1,fmt=*) vertices(1,2)
               gnu_string = trim(gnu_string) // trim(str_temp1) // ","
               write(str_temp1,fmt=*) vertices(2,2)
               gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
               write(str_temp1,fmt=*) vertices(1,3)
               gnu_string = trim(gnu_string) // trim(str_temp1) // ","
               write(str_temp1,fmt=*) vertices(2,3)
               gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
               write(str_temp1,fmt=*) vertices(1,4)
               gnu_string = trim(gnu_string) // trim(str_temp1) // ","
               write(str_temp1,fmt=*) vertices(2,4)
               gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
               write(str_temp1,fmt=*) vertices(1,1)
               gnu_string = trim(gnu_string) // trim(str_temp1) // ","
               write(str_temp1,fmt=*) vertices(2,1)
               gnu_string = trim(gnu_string) // trim(str_temp1)
               gnu_string = trim(gnu_string) // " fillcolor rgb '#B860CD' fillstyle solid noborder"

               if (i == 0) then
                    gnu_string = trim(gnu_string) // "/"
               end if
          end do
     else
          height = 0.025_dp

          side = height / sin( acos(-1.0_dp) / 3.0_dp )
          radius = side / sqrt(3.0_dp)
          apothem = height - radius

          if (info_graph == 3 .or. info_graph == 4) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),5)
          else if (info_graph == 6) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),7)
          else if (info_graph == 8) then
               point_coord(1) = interval_shift(point_coord(1),info_graph)
               point_coord(2) = interval_shift(point_coord(2),9)
          else if (info_graph == 11 .or. info_graph == 12) then
               point_coord(1) = interval_shift(point_coord(1),10)
               point_coord(2) = interval_shift(point_coord(2),info_graph)
          else
               point_coord(1) = interval_shift(point_coord(1),info_graph)
          end if

          vertices(:,1) = (/point_coord(1) - (side / 2.0_dp), point_coord(2) - apothem/)
          vertices(:,2) = (/point_coord(1) + (side / 2.0_dp), point_coord(2) - apothem/)
          vertices(:,3) = (/point_coord(1), point_coord(2) + radius/)

          gnu_string = ""
          gnu_string = trim(gnu_string) // "set object polygon front from graph "
          write(str_temp1,fmt=*) vertices(1,1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
          write(str_temp1,fmt=*) vertices(1,2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,2)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
          write(str_temp1,fmt=*) vertices(1,3)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,3)
          gnu_string = trim(gnu_string) // trim(str_temp1) // " to graph "
          write(str_temp1,fmt=*) vertices(1,1)
          gnu_string = trim(gnu_string) // trim(str_temp1) // ","
          write(str_temp1,fmt=*) vertices(2,1)
          gnu_string = trim(gnu_string) // trim(str_temp1)
          gnu_string = trim(gnu_string) // " fillcolor rgb 'red' fillstyle solid noborder"
     end if

end subroutine marker

!                                                                                                                    :::::::::::::::::::::::::::::
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

end module m_subrutinas
