program modelo_HH

! ::: PROGRAMA: MODELO DE HODGKIN-HUXLEY :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! 
!      Este fichero contiene el programa para la integración numérica del modelo de Hodgkin-Huxley, tanto la versión original 4D como la versión
!   reducida 2D. Además también podrán calcularse las funciones de activación/inactivación y las constantes de tiempo. El propio programa
!   presentará al usuario durante su ejecución las opciones disponibles.
! 
!      Para la integración del modelo el programa tiene en cuenta el valor del parámetro de bifurcación E_K, pidiéndole de antemano su valor al 
!   usuario. La estimulación eléctrica a la neurona se realiza mediante una corriente externa. Para modificar dicha estimulación basta con
!   cambiar en el programa los parámetros que se pasan a la función 'HH_estimulo(t,I_externa,t_inicial,t_final)' (destacada en el código):
!   
!           - t (variable real)                   : Representa el tiempo (ms)
!           - I_externa (variable real)           : Recoge el valor de la corriente externa usada para el estímulo (μA)
!           - t_inicial (variable real, opcional) : Recoge el tiempo de inicio del estímulo (ms)
!           - t_final (variable real, opcional)   : Recoge el tiempo de fin del estímulo (ms)
!   
!      Las variables t_inicial y t_final actúan como extremos del intervalo de actuación del estímulo. Ante la ausencia de t_inicial (t_final), la
!   función coloca el extremo del intervalo al inicio (final) del tiempo de integración de las ecuaciones diferenciales.
!   
!      Para la determinación de las condiciones iniciales se asume la situación considerada en el trabajo; se parte de la neurona en reposo y, 
!   sin ningún tipo de estímulo externo, se aumenta E_K. Así, el programa supone como condiciones iniciales el estado en el que se encontraría
!   la neurona para un valor determinado de E_K (proporcionado por el usuario). El usuario podrá especificar un valor de E_K en el intervalo
!   [-77,-40] mV y con una precisión máxima de 3 decimales.
!
! ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

use m_parametros
use m_modelo_HH_4D
use m_modelo_HH_2D
use m_subrutinas

real(kind=dp) :: t, V_mem
real(kind=dp), dimension(:), allocatable :: y

real(kind=dp) :: TMAX = 200.0_dp ! ms
real(kind=dp) :: h = 0.001_dp

integer :: IOSTAT

print *
print *, "-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_"
print *, "-_                                                                                                                -_"
print *, "-_   MODELO DE HODGKIN-HUXLEY                                                                                     -_"
print *, "-_   ------------------------                                                                                     -_"
print *, "-_                                                                                                                -_"
print *, "-_      El cálculo tanto de las soluciones del modelo de Hodgkin-Huxley como de sus cualidades principales        -_"
print *, "-_   puede hacerse sobre el modelo 4D (original) o el modelo 2D (reducido). Indicar una acción con un número      -_"
print *, "-_   entero:                                                                                                      -_"
print *, "-_                                                                                                                -_"
print *, "-_      (1) Modelo de Hodgkin-Huxley original                                                                     -_"
print *, "-_            - Integración del modelo original de Hodgkin-Huxley                                                 -_"
print *, "-_            - Funciones de activación/inactivación y constantes de tiempo                                       -_"
print *, "-_      (2) Modelo de Hodgkin-Huxley reducido                                                                     -_"
print *, "-_            - Integración del modelo reducido de Hodgkin-Huxley                                                 -_"
print *, "-_            - Función de activación y constante de tiempo de K+                                                 -_"
print *, "-_                                                                                                                -_"
print *, "-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_"
print *
read  *, IOSTAT

select case(IOSTAT)

case(1)

     allocate(y(4))

     print *
     print *, "--------------------------------------------------------------------------------------------------------------------"
     print *, "--------------------------------------------------------------------------------------------------------------------"
     print *, "     MODELO DE HODGKIN-HUXLEY ORIGINAL (4D)"
     print *, "--------------------------------------------------------------------------------------------------------------------"
     print *
     print *, "   Indicar una acción:"
     print *, "      (1) Integración del modelo 4D de Hodgkin-Huxley"
     print *, "      (2) Funciones de activación/inactivación y constantes de tiempo"
     print *
     read  *, IOSTAT

     IOSTAT = IOSTAT + 2

     select case(IOSTAT)

     case(3)

          print *
          print *, "   Integración del modelo 4D de Hodgkin-Huxley"
          print *, "   ----------------------------------------------------------------"
          print *
          print *, "   Introducir el valor del parámetro de bifurcación E_K (en mV):"
          print *
          read  *, E_K
          print *
          print *
          print *
          print *, "   Calculando..."

          ! --- Cálculo de condiciones iniciales: ---
          t = 0.0_dp
          call HH_cond_iniciales_4D(E_K,y)

          ! --- Integración de las ec. diferenciales: ---
          open(unit=1, file="modelo_HH_4D.txt", status="replace", action="write")
          write(unit=1,fmt=*) t, y(1), y(2), y(3), y(4)
          do
               if (t > TMAX) then
                    exit
               end if
     ! --- Modificación del estímulo eléctrico: --------------------------------------------------------
     ! -------------------------------------------------------------------------------------------------
     ! ----------------
               call HH_estimulo(t=t,I_externa=10.0_dp,t_inicial=100.0_dp,t_final=101.0_dp)
!               call HH_estimulo(t=t,I_externa=0.0_dp)
     !                                                                                  ----------------
     ! -------------------------------------------------------------------------------------------------
     ! -------------------------------------------------------------------------------------------------
               call rk4(t,y,HH_modelo_4D,h)
               write(unit=1,fmt=*) t, y(1), y(2), y(3), y(4)
          end do
          close(unit=1)

          ! --- Creación de la gráfica: ---
          call gnu_file(IOSTAT)
          call system("gnuplot gnu_temp.gnu")

          call system("rm gnu_temp.gnu")

          print *
          print *
          print *
          print *, "   --- Gráfica: ---"
          print *, "   --------------------------------------------"
          print *, "      Output:"
          print *, "         'modelo_HH_4D.txt'"
          print *, "         'modelo_HH_4D.pdf'"
          print *
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *

     case(4)

          print *
          print *, "   Funciones de activación/inactivación y constantes de tiempo"
          print *, "   ----------------------------------------------------------------"
          print *
          print *
          print *
          print *, "   Calculando..."

          ! --- Cálculos para las gráficas: ---
          open(unit=1, file="modelo_HH_4D_caracteristicas.txt", status="replace", action="write")
          V_mem = -100.0_dp
          do
               if (V_mem > 80.0_dp) then
                    exit
               end if
               call HH_funciones_act_inact_4D(V_mem,y)
               write(unit=1,fmt=*) V_mem, y(2), y(3), y(4)
               V_mem = V_mem + 0.001_dp
          end do
          write(unit=1,fmt=*)
          write(unit=1,fmt=*)

          V_mem = -100.0_dp
          do
               if (V_mem > 80.0_dp) then
                    exit
               end if
               call HH_ctes_tiempo_4D(V_mem,y)
               write(unit=1,fmt=*) V_mem, y(2), y(3), y(4)
               V_mem = V_mem + 0.001_dp
          end do
          close(unit=1)

          ! --- Creación de las gráficas: ---
          call gnu_file(IOSTAT)
          call system("gnuplot gnu_temp.gnu")

          call system("rm gnu_temp.gnu")

          print *
          print *
          print *
          print *, "   --- Gráficas: ---"
          print *, "   --------------------------------------------"
          print *, "      Output:"
          print *, "         'modelo_HH_4D_caracteristicas.txt'"
          print *, "         'func_act_inact_4D.pdf', 'ctes_tiempo_4D.pdf'"
          print *
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *

     case default

          print *
          print *, "* * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR IOSTAT: Acción no válida     *"
          print *, "* * * * * * * * * * * * * * * * * * * *"
          print *

     end select

     deallocate(y)

case(2)

     allocate(y(2))

     print *
     print *, "--------------------------------------------------------------------------------------------------------------------"
     print *, "--------------------------------------------------------------------------------------------------------------------"
     print *, "     MODELO DE HODGKIN-HUXLEY REDUCIDO (2D)"
     print *, "--------------------------------------------------------------------------------------------------------------------"
     print *
     print *, "     Indicar una acción:"
     print *, "        (1) Integración del modelo 2D de Hodgkin-Huxley"
     print *, "        (2) Función de activación y constante de tiempo de K+"
     print *
     read  *, IOSTAT

     IOSTAT = IOSTAT + 4

     select case(IOSTAT)

     case(5)

          print *
          print *, "   Integración del modelo 2D de Hodgkin-Huxley"
          print *, "   ----------------------------------------------------------------"
          print *
          print *, "   Introducir el valor del parámetro de bifurcación E_K (en mV):"
          print *
          read  *, E_K
          print *
          print *
          print *
          print *, "   Calculando..."

          ! --- Cálculo de condiciones iniciales: ---
          t = 0.0_dp
          call HH_cond_iniciales_2D(E_K,y)

          ! --- Integración de las ec. diferenciales: ---
          open(unit=1, file="modelo_HH_2D.txt", status="replace", action="write")
          write(unit=1,fmt=*) t, y(1), y(2)
          do
               if (t > TMAX) then
                    exit
               end if
     ! --- Modificación del estímulo eléctrico: --------------------------------------------------------
     ! -------------------------------------------------------------------------------------------------
     ! ----------------
               call HH_estimulo(t=t,I_externa=10.0_dp,t_inicial=100.0_dp,t_final=101.0_dp)
!               call HH_estimulo(t=t,I_externa=0.0_dp)

     !                                                                                  ----------------
     ! -------------------------------------------------------------------------------------------------
     ! -------------------------------------------------------------------------------------------------
               call rk4(t,y,HH_modelo_2D,h)
               write(unit=1,fmt=*) t, y(1), y(2)
          end do
          close(unit=1)

          ! --- Creación de la gráfica: ---
          call gnu_file(IOSTAT)
          call system("gnuplot gnu_temp.gnu")

          call system("rm gnu_temp.gnu")

          print *
          print *
          print *
          print *, "   --- Gráfica: ---"
          print *, "   --------------------------------------------"
          print *, "      Output:"
          print *, "         'modelo_HH_2D.txt'"
          print *, "         'modelo_HH_2D.pdf'"
          print *
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *

     case(6)

          print *
          print *, "   Función de activación y constante de tiempo de K+"
          print *, "   ----------------------------------------------------------------"
          print *
          print *
          print *
          print *, "   Calculando..."

          ! --- Cálculos para las gráficas: ---
          open(unit=1, file="modelo_HH_2D_caracteristicas.txt", status="replace", action="write")
          V_mem = -100.0_dp
          do
               if (V_mem > 80.0_dp) then
                    exit
               end if
               call HH_funciones_act_inact_2D(V_mem,y)
               write(unit=1,fmt=*) V_mem, y(2)
               V_mem = V_mem + 0.001_dp
          end do
          write(unit=1,fmt=*)
          write(unit=1,fmt=*)

          V_mem = -100.0_dp
          do
               if (V_mem > 80.0_dp) then
                    exit
               end if
               call HH_ctes_tiempo_2D(V_mem,y)
               write(unit=1,fmt=*) V_mem, y(2)
               V_mem = V_mem + 0.001_dp
          end do
          close(unit=1)

          ! --- Creación de las gráficas: ---
          call gnu_file(IOSTAT)
          call system("gnuplot gnu_temp.gnu")

          call system("rm gnu_temp.gnu")

          print *
          print *
          print *
          print *, "   --- Gráficas: ---"
          print *, "   --------------------------------------------"
          print *, "      Output:"
          print *, "         'modelo_HH_2D_caracteristicas.txt'"
          print *, "         'func_act_2D.pdf', 'cte_tiempo_2D.pdf'"
          print *
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *, "----------------------------------------------------------------------------------------------------------&
                   &----------"
          print *

     case default

          print *
          print *, "* * * * * * * * * * * * * * * * * * * *"
          print *, "*  ERROR IOSTAT: Acción no válida     *"
          print *, "* * * * * * * * * * * * * * * * * * * *"
          print *

     end select

     deallocate(y)

case default

     print *
     print *, "* * * * * * * * * * * * * * * * * * * *"
     print *, "*  ERROR IOSTAT: Acción no válida     *"
     print *, "* * * * * * * * * * * * * * * * * * * *"
     print *

end select

end program modelo_HH
